# Hidden Markov Models
#
# Author: Ron Weiss <ronweiss@gmail.com>
#         Shiqiao Du <lucidfrontier.45@gmail.com>
# API changes: Jaques Grobler <jaquesgrobler@gmail.com>
# Modifications to create of the HMMLearn module: Gael Varoquaux
# More API changes: Sergei Lebedev <superbobry@gmail.com>

"""
The :mod:`hmmlearn.hmm` module implements hidden Markov models.
"""

import logging

import numpy as np
from scipy.special import logsumexp
from sklearn import cluster
from sklearn.utils import check_random_state

from . import _utils
from .stats import log_multivariate_normal_density
from .base import _BaseHMM
from .utils import (
    fill_covars, iter_from_X_lengths, log_mask_zero, log_normalize, normalize)

__all__ = ["GMMHMM", "GaussianHMM", "MultinomialHMM"]


_log = logging.getLogger(__name__)
COVARIANCE_TYPES = frozenset(("spherical", "diag", "full", "tied"))


def _check_and_set_gaussian_n_features(model, X):
    _, n_features = X.shape
    if hasattr(model, "n_features") and model.n_features != n_features:
        raise ValueError("Unexpected number of dimensions, got {} but "
                         "expected {}".format(n_features, model.n_features))
    model.n_features = n_features


class GaussianHMM(_BaseHMM):
    r"""Hidden Markov Model with Gaussian emissions.

    Parameters
    ----------
    n_components : int
        Number of states.

    covariance_type : string, optional
        String describing the type of covariance parameters to
        use.  Must be one of

        * "spherical" --- each state uses a single variance value that
          applies to all features.
        * "diag" --- each state uses a diagonal covariance matrix.
        * "full" --- each state uses a full (i.e. unrestricted)
          covariance matrix.
        * "tied" --- all states use **the same** full covariance matrix.

        Defaults to "diag".

    min_covar : float, optional
        Floor on the diagonal of the covariance matrix to prevent
        overfitting. Defaults to 1e-3.

    startprob_prior : array, shape (n_components, ), optional
        Parameters of the Dirichlet prior distribution for
        :attr:`startprob_`.

    transmat_prior : array, shape (n_components, n_components), optional
        Parameters of the Dirichlet prior distribution for each row
        of the transition probabilities :attr:`transmat_`.

    means_prior, means_weight : array, shape (n_components, ), optional
        Mean and precision of the Normal prior distribtion for
        :attr:`means_`.

    covars_prior, covars_weight : array, shape (n_components, ), optional
        Parameters of the prior distribution for the covariance matrix
        :attr:`covars_`.

        If :attr:`covariance_type` is "spherical" or "diag" the prior is
        the inverse gamma distribution, otherwise --- the inverse Wishart
        distribution.

    algorithm : string, optional
        Decoder algorithm. Must be one of "viterbi" or`"map".
        Defaults to "viterbi".

    random_state: RandomState or an int seed, optional
        A random number generator instance.

    n_iter : int, optional
        Maximum number of iterations to perform.

    tol : float, optional
        Convergence threshold. EM will stop if the gain in log-likelihood
        is below this value.

    verbose : bool, optional
        When ``True`` per-iteration convergence reports are printed
        to :data:`sys.stderr`. You can diagnose convergence via the
        :attr:`monitor_` attribute.

    params : string, optional
        Controls which parameters are updated in the training
        process.  Can contain any combination of 's' for startprob,
        't' for transmat, 'm' for means and 'c' for covars. Defaults
        to all parameters.

    init_params : string, optional
        Controls which parameters are initialized prior to
        training.  Can contain any combination of 's' for
        startprob, 't' for transmat, 'm' for means and 'c' for covars.
        Defaults to all parameters.

    Attributes
    ----------
    n_features : int
        Dimensionality of the Gaussian emissions.

    monitor\_ : ConvergenceMonitor
        Monitor object used to check the convergence of EM.

    startprob\_ : array, shape (n_components, )
        Initial state occupation distribution.

    transmat\_ : array, shape (n_components, n_components)
        Matrix of transition probabilities between states.

    means\_ : array, shape (n_components, n_features)
        Mean parameters for each state.

    covars\_ : array
        Covariance parameters for each state.

        The shape depends on :attr:`covariance_type`::

            (n_components, )                        if "spherical",
            (n_components, n_features)              if "diag",
            (n_components, n_features, n_features)  if "full"
            (n_features, n_features)                if "tied",

    Examples
    --------
    >>> from hmmlearn.hmm import GaussianHMM
    >>> GaussianHMM(n_components=2)  #doctest: +ELLIPSIS
    GaussianHMM(algorithm='viterbi',...
    """
    def __init__(self, n_components=1, covariance_type='diag',
                 min_covar=1e-3,
                 startprob_prior=1.0, transmat_prior=1.0,
                 means_prior=0, means_weight=0,
                 covars_prior=1e-2, covars_weight=1,
                 algorithm="viterbi", random_state=None,
                 n_iter=10, tol=1e-2, verbose=False,
                 params="stmc", init_params="stmc"):
        _BaseHMM.__init__(self, n_components,
                          startprob_prior=startprob_prior,
                          transmat_prior=transmat_prior, algorithm=algorithm,
                          random_state=random_state, n_iter=n_iter,
                          tol=tol, params=params, verbose=verbose,
                          init_params=init_params)

        self.covariance_type = covariance_type
        self.min_covar = min_covar
        self.means_prior = means_prior
        self.means_weight = means_weight
        self.covars_prior = covars_prior
        self.covars_weight = covars_weight

    @property
    def covars_(self):
        """Return covars as a full matrix."""
        return fill_covars(self._covars_, self.covariance_type,
                           self.n_components, self.n_features)

    @covars_.setter
    def covars_(self, covars):
        covars = np.array(covars, copy=True)
        _utils._validate_covars(covars, self.covariance_type,
                                self.n_components)
        self._covars_ = covars

    def _get_n_fit_scalars_per_param(self):
        nc = self.n_components
        nf = self.n_features
        return {
            "s": nc - 1,
            "t": nc * (nc - 1),
            "m": nc * nf,
            "c": {
                "spherical": nc,
                "diag": nc * nf,
                "full": nc * nf * (nf + 1) // 2,
                "tied": nf * (nf + 1) // 2,
            }[self.covariance_type],
        }

    def _init(self, X, lengths=None):
        _check_and_set_gaussian_n_features(self, X)
        super()._init(X, lengths=lengths)

        if self._needs_init("m", "means_"):
            kmeans = cluster.KMeans(n_clusters=self.n_components,
                                    random_state=self.random_state)
            kmeans.fit(X)
            self.means_ = kmeans.cluster_centers_
        if self._needs_init("c", "covars_"):
            cv = np.cov(X.T) + self.min_covar * np.eye(X.shape[1])
            if not cv.shape:
                cv.shape = (1, 1)
            self.covars_ = \
                _utils.distribute_covar_matrix_to_match_covariance_type(
                    cv, self.covariance_type, self.n_components).copy()

    def _check(self):
        super()._check()

        self.means_ = np.asarray(self.means_)
        self.n_features = self.means_.shape[1]

        if self.covariance_type not in COVARIANCE_TYPES:
            raise ValueError('covariance_type must be one of {}'
                             .format(COVARIANCE_TYPES))

    def _compute_log_likelihood(self, X):
        return log_multivariate_normal_density(
            X, self.means_, self._covars_, self.covariance_type)

    def _generate_sample_from_state(self, state, random_state=None):
        random_state = check_random_state(random_state)
        return random_state.multivariate_normal(
            self.means_[state], self.covars_[state]
        )

    def _initialize_sufficient_statistics(self):
        stats = super()._initialize_sufficient_statistics()
        stats['post'] = np.zeros(self.n_components)
        stats['obs'] = np.zeros((self.n_components, self.n_features))
        stats['obs**2'] = np.zeros((self.n_components, self.n_features))
        if self.covariance_type in ('tied', 'full'):
            stats['obs*obs.T'] = np.zeros((self.n_components, self.n_features,
                                           self.n_features))
        return stats

    def _accumulate_sufficient_statistics(self, stats, obs, framelogprob,
                                          posteriors, fwdlattice, bwdlattice):
        super()._accumulate_sufficient_statistics(
            stats, obs, framelogprob, posteriors, fwdlattice, bwdlattice)

        if 'm' in self.params or 'c' in self.params:
            stats['post'] += posteriors.sum(axis=0)
            stats['obs'] += np.dot(posteriors.T, obs)

        if 'c' in self.params:
            if self.covariance_type in ('spherical', 'diag'):
                stats['obs**2'] += np.dot(posteriors.T, obs ** 2)
            elif self.covariance_type in ('tied', 'full'):
                # posteriors: (nt, nc); obs: (nt, nf); obs: (nt, nf)
                # -> (nc, nf, nf)
                stats['obs*obs.T'] += np.einsum(
                    'ij,ik,il->jkl', posteriors, obs, obs)

    def _do_mstep(self, stats):
        super()._do_mstep(stats)

        means_prior = self.means_prior
        means_weight = self.means_weight

        # TODO: find a proper reference for estimates for different
        #       covariance models.
        # Based on Huang, Acero, Hon, "Spoken Language Processing",
        # p. 443 - 445
        denom = stats['post'][:, None]
        if 'm' in self.params:
            self.means_ = ((means_weight * means_prior + stats['obs'])
                           / (means_weight + denom))

        if 'c' in self.params:
            covars_prior = self.covars_prior
            covars_weight = self.covars_weight
            meandiff = self.means_ - means_prior

            if self.covariance_type in ('spherical', 'diag'):
                cv_num = (means_weight * meandiff**2
                          + stats['obs**2']
                          - 2 * self.means_ * stats['obs']
                          + self.means_**2 * denom)
                cv_den = max(covars_weight - 1, 0) + denom
                self._covars_ = \
                    (covars_prior + cv_num) / np.maximum(cv_den, 1e-5)
                if self.covariance_type == 'spherical':
                    self._covars_ = np.tile(self._covars_.mean(1)[:, None],
                                            (1, self._covars_.shape[1]))
            elif self.covariance_type in ('tied', 'full'):
                cv_num = np.empty((self.n_components, self.n_features,
                                  self.n_features))
                for c in range(self.n_components):
                    obsmean = np.outer(stats['obs'][c], self.means_[c])

                    cv_num[c] = (means_weight * np.outer(meandiff[c],
                                                         meandiff[c])
                                 + stats['obs*obs.T'][c]
                                 - obsmean - obsmean.T
                                 + np.outer(self.means_[c], self.means_[c])
                                 * stats['post'][c])
                cvweight = max(covars_weight - self.n_features, 0)
                if self.covariance_type == 'tied':
                    self._covars_ = ((covars_prior + cv_num.sum(axis=0)) /
                                     (cvweight + stats['post'].sum()))
                elif self.covariance_type == 'full':
                    self._covars_ = ((covars_prior + cv_num) /
                                     (cvweight + stats['post'][:, None, None]))


class MultinomialHMM(_BaseHMM):
    r"""Hidden Markov Model with multinomial (discrete) emissions

    Parameters
    ----------

    n_components : int
        Number of states.

    startprob_prior : array, shape (n_components, ), optional
        Parameters of the Dirichlet prior distribution for
        :attr:`startprob_`.

    transmat_prior : array, shape (n_components, n_components), optional
        Parameters of the Dirichlet prior distribution for each row
        of the transition probabilities :attr:`transmat_`.

    algorithm : string, optional
        Decoder algorithm. Must be one of "viterbi" or "map".
        Defaults to "viterbi".

    random_state: RandomState or an int seed, optional
        A random number generator instance.

    n_iter : int, optional
        Maximum number of iterations to perform.

    tol : float, optional
        Convergence threshold. EM will stop if the gain in log-likelihood
        is below this value.

    verbose : bool, optional
        When ``True`` per-iteration convergence reports are printed
        to :data:`sys.stderr`. You can diagnose convergence via the
        :attr:`monitor_` attribute.

    params : string, optional
        Controls which parameters are updated in the training
        process.  Can contain any combination of 's' for startprob,
        't' for transmat, 'e' for emissionprob.
        Defaults to all parameters.

    init_params : string, optional
        Controls which parameters are initialized prior to
        training.  Can contain any combination of 's' for
        startprob, 't' for transmat, 'e' for emissionprob.
        Defaults to all parameters.

    Attributes
    ----------
    n_features : int
        Number of possible symbols emitted by the model (in the samples).

    monitor\_ : ConvergenceMonitor
        Monitor object used to check the convergence of EM.

    startprob\_ : array, shape (n_components, )
        Initial state occupation distribution.

    transmat\_ : array, shape (n_components, n_components)
        Matrix of transition probabilities between states.

    emissionprob\_ : array, shape (n_components, n_features)
        Probability of emitting a given symbol when in each state.

    Examples
    --------
    >>> from hmmlearn.hmm import MultinomialHMM
    >>> MultinomialHMM(n_components=2)  #doctest: +ELLIPSIS
    MultinomialHMM(algorithm='viterbi',...
    """
    # TODO: accept the prior on emissionprob_ for consistency.
    def __init__(self, n_components=1,
                 startprob_prior=1.0, transmat_prior=1.0,
                 algorithm="viterbi", random_state=None,
                 n_iter=10, tol=1e-2, verbose=False,
                 params="ste", init_params="ste"):
        _BaseHMM.__init__(self, n_components,
                          startprob_prior=startprob_prior,
                          transmat_prior=transmat_prior,
                          algorithm=algorithm,
                          random_state=random_state,
                          n_iter=n_iter, tol=tol, verbose=verbose,
                          params=params, init_params=init_params)

    def _get_n_fit_scalars_per_param(self):
        nc = self.n_components
        nf = self.n_features
        return {
            "s": nc - 1,
            "t": nc * (nc - 1),
            "e": nc * (nf - 1),
        }

    def _init(self, X, lengths=None):
        self._check_and_set_n_features(X)
        super()._init(X, lengths=lengths)
        self.random_state = check_random_state(self.random_state)

        if 'e' in self.init_params:
            self.emissionprob_ = self.random_state \
                .rand(self.n_components, self.n_features)
            normalize(self.emissionprob_, axis=1)

    def _check(self):
        super()._check()

        self.emissionprob_ = np.atleast_2d(self.emissionprob_)
        n_features = getattr(self, "n_features", self.emissionprob_.shape[1])
        if self.emissionprob_.shape != (self.n_components, n_features):
            raise ValueError(
                "emissionprob_ must have shape (n_components, n_features)")
        else:
            self.n_features = n_features

    def _compute_log_likelihood(self, X):
        return log_mask_zero(self.emissionprob_)[:, np.concatenate(X)].T

    def _generate_sample_from_state(self, state, random_state=None):
        cdf = np.cumsum(self.emissionprob_[state, :])
        random_state = check_random_state(random_state)
        return [(cdf > random_state.rand()).argmax()]

    def _initialize_sufficient_statistics(self):
        stats = super()._initialize_sufficient_statistics()
        stats['obs'] = np.zeros((self.n_components, self.n_features))
        return stats

    def _accumulate_sufficient_statistics(self, stats, X, framelogprob,
                                          posteriors, fwdlattice, bwdlattice):
        super()._accumulate_sufficient_statistics(
            stats, X, framelogprob, posteriors, fwdlattice, bwdlattice)
        if 'e' in self.params:
            for t, symbol in enumerate(np.concatenate(X)):
                stats['obs'][:, symbol] += posteriors[t]

    def _do_mstep(self, stats):
        super()._do_mstep(stats)
        if 'e' in self.params:
            self.emissionprob_ = (
                stats['obs'] / stats['obs'].sum(axis=1, keepdims=True))

    def _check_and_set_n_features(self, X):
        """
        Check if ``X`` is a sample from a Multinomial distribution, i.e. an
        array of non-negative integers.
        """
        if not np.issubdtype(X.dtype, np.integer):
            raise ValueError("Symbols should be integers")
        if X.min() < 0:
            raise ValueError("Symbols should be nonnegative")
        if hasattr(self, "n_features"):
            if self.n_features - 1 < X.max():
                raise ValueError(
                    "Largest symbol is {} but the model only emits "
                    "symbols up to {}"
                    .format(X.max(), self.n_features - 1))
        self.n_features = X.max() + 1


class GMMHMM(_BaseHMM):
    r"""Hidden Markov Model with Gaussian mixture emissions.

    Parameters
    ----------
    n_components : int
        Number of states in the model.

    n_mix : int
        Number of states in the GMM.

    covariance_type : string, optional
        String describing the type of covariance parameters to
        use.  Must be one of

        * "spherical" --- each state uses a single variance value that
          applies to all features.
        * "diag" --- each state uses a diagonal covariance matrix.
        * "full" --- each state uses a full (i.e. unrestricted)
          covariance matrix.
        * "tied" --- all mixture components of each state use **the same** full
          covariance matrix (note that this is not the same as for
          `GaussianHMM`).

        Defaults to "diag".

    min_covar : float, optional
        Floor on the diagonal of the covariance matrix to prevent
        overfitting. Defaults to 1e-3.

    startprob_prior : array, shape (n_components, ), optional
        Parameters of the Dirichlet prior distribution for
        :attr:`startprob_`.

    transmat_prior : array, shape (n_components, n_components), optional
        Parameters of the Dirichlet prior distribution for each row
        of the transition probabilities :attr:`transmat_`.

    weights_prior : array, shape (n_mix, ), optional
        Parameters of the Dirichlet prior distribution for
        :attr:`weights_`.

    means_prior, means_weight : array, shape (n_mix, ), optional
        Mean and precision of the Normal prior distribtion for
        :attr:`means_`.

    covars_prior, covars_weight : array, shape (n_mix, ), optional
        Parameters of the prior distribution for the covariance matrix
        :attr:`covars_`.

        If :attr:`covariance_type` is "spherical" or "diag" the prior is
        the inverse gamma distribution, otherwise --- the inverse Wishart
        distribution.

    algorithm : string, optional
        Decoder algorithm. Must be one of "viterbi" or "map".
        Defaults to "viterbi".

    random_state: RandomState or an int seed, optional
        A random number generator instance.

    n_iter : int, optional
        Maximum number of iterations to perform.

    tol : float, optional
        Convergence threshold. EM will stop if the gain in log-likelihood
        is below this value.

    verbose : bool, optional
        When ``True`` per-iteration convergence reports are printed
        to :data:`sys.stderr`. You can diagnose convergence via the
        :attr:`monitor_` attribute.

    init_params : string, optional
        Controls which parameters are initialized prior to training. Can
        contain any combination of 's' for startprob, 't' for transmat, 'm'
        for means, 'c' for covars, and 'w' for GMM mixing weights.
        Defaults to all parameters.

    params : string, optional
        Controls which parameters are updated in the training process.  Can
        contain any combination of 's' for startprob, 't' for transmat, 'm' for
        means, and 'c' for covars, and 'w' for GMM mixing weights.
        Defaults to all parameters.

    Attributes
    ----------
    monitor\_ : ConvergenceMonitor
        Monitor object used to check the convergence of EM.

    startprob\_ : array, shape (n_components, )
        Initial state occupation distribution.

    transmat\_ : array, shape (n_components, n_components)
        Matrix of transition probabilities between states.

    weights\_ : array, shape (n_components, n_mix)
        Mixture weights for each state.

    means\_ : array, shape (n_components, n_mix)
        Mean parameters for each mixture component in each state.

    covars\_ : array
        Covariance parameters for each mixture components in each state.

        The shape depends on :attr:`covariance_type`::

            (n_components, n_mix)                          if "spherical",
            (n_components, n_mix, n_features)              if "diag",
            (n_components, n_mix, n_features, n_features)  if "full"
            (n_components, n_features, n_features)         if "tied",
    """

    def __init__(self, n_components=1, n_mix=1,
                 min_covar=1e-3, startprob_prior=1.0, transmat_prior=1.0,
                 weights_prior=1.0, means_prior=0.0, means_weight=0.0,
                 covars_prior=None, covars_weight=None,
                 algorithm="viterbi", covariance_type="diag",
                 random_state=None, n_iter=10, tol=1e-2,
                 verbose=False, params="stmcw",
                 init_params="stmcw"):
        _BaseHMM.__init__(self, n_components,
                          startprob_prior=startprob_prior,
                          transmat_prior=transmat_prior,
                          algorithm=algorithm, random_state=random_state,
                          n_iter=n_iter, tol=tol, verbose=verbose,
                          params=params, init_params=init_params)
        self.covariance_type = covariance_type
        self.min_covar = min_covar
        self.n_mix = n_mix
        self.weights_prior = weights_prior
        self.means_prior = means_prior
        self.means_weight = means_weight
        self.covars_prior = covars_prior
        self.covars_weight = covars_weight

    def _get_n_fit_scalars_per_param(self):
        nc = self.n_components
        nf = self.n_features
        nm = self.n_mix
        return {
            "s": nc - 1,
            "t": nc * (nc - 1),
            "m": nc * nm * nf,
            "c": {
                "spherical": nc * nm,
                "diag": nc * nm * nf,
                "full": nc * nm * nf * (nf + 1) // 2,
                "tied": nc * nf * (nf + 1) // 2,
            }[self.covariance_type],
            "w": nm - 1,
        }

    def _init(self, X, lengths=None):
        _check_and_set_gaussian_n_features(self, X)
        super()._init(X, lengths=lengths)
        nc = self.n_components
        nf = self.n_features
        nm = self.n_mix

        # Default values for covariance prior parameters
        self._init_covar_priors()
        self._fix_priors_shape()

        main_kmeans = cluster.KMeans(n_clusters=nc,
                                     random_state=self.random_state)
        labels = main_kmeans.fit_predict(X)
        kmeanses = []
        for label in range(nc):
            kmeans = cluster.KMeans(n_clusters=nm,
                                    random_state=self.random_state)
            kmeans.fit(X[np.where(labels == label)])
            kmeanses.append(kmeans)

        if self._needs_init("w", "weights_"):
            self.weights_ = np.ones((nc, nm)) / (np.ones((nc, 1)) * nm)

        if self._needs_init("m", "means_"):
            self.means_ = np.stack(
                [kmeans.cluster_centers_ for kmeans in kmeanses])

        if self._needs_init("c", "covars_"):
            cv = np.cov(X.T) + self.min_covar * np.eye(nf)
            if not cv.shape:
                cv.shape = (1, 1)
            if self.covariance_type == 'tied':
                self.covars_ = np.zeros((nc, nf, nf))
                self.covars_[:] = cv
            elif self.covariance_type == 'full':
                self.covars_ = np.zeros((nc, nm, nf, nf))
                self.covars_[:] = cv
            elif self.covariance_type == 'diag':
                self.covars_ = np.zeros((nc, nm, nf))
                self.covars_[:] = np.diag(cv)
            elif self.covariance_type == 'spherical':
                self.covars_ = np.zeros((nc, nm))
                self.covars_[:] = cv.mean()

    def _init_covar_priors(self):
        if self.covariance_type == "full":
            if self.covars_prior is None:
                self.covars_prior = 0.0
            if self.covars_weight is None:
                self.covars_weight = -(1.0 + self.n_features + 1.0)
        elif self.covariance_type == "tied":
            if self.covars_prior is None:
                self.covars_prior = 0.0
            if self.covars_weight is None:
                self.covars_weight = -(self.n_mix + self.n_features + 1.0)
        elif self.covariance_type == "diag":
            if self.covars_prior is None:
                self.covars_prior = -1.5
            if self.covars_weight is None:
                self.covars_weight = 0.0
        elif self.covariance_type == "spherical":
            if self.covars_prior is None:
                self.covars_prior = -(self.n_mix + 2.0) / 2.0
            if self.covars_weight is None:
                self.covars_weight = 0.0

    def _fix_priors_shape(self):
        nc = self.n_components
        nf = self.n_features
        nm = self.n_mix

        # If priors are numbers, this function will make them into a
        # matrix of proper shape
        self.weights_prior = np.broadcast_to(
            self.weights_prior, (nc, nm)).copy()
        self.means_prior = np.broadcast_to(
            self.means_prior, (nc, nm, nf)).copy()
        self.means_weight = np.broadcast_to(
            self.means_weight, (nc, nm)).copy()

        if self.covariance_type == "full":
            self.covars_prior = np.broadcast_to(
                self.covars_prior, (nc, nm, nf, nf)).copy()
            self.covars_weight = np.broadcast_to(
                self.covars_weight, (nc, nm)).copy()
        elif self.covariance_type == "tied":
            self.covars_prior = np.broadcast_to(
                self.covars_prior, (nc, nf, nf)).copy()
            self.covars_weight = np.broadcast_to(
                self.covars_weight, nc).copy()
        elif self.covariance_type == "diag":
            self.covars_prior = np.broadcast_to(
                self.covars_prior, (nc, nm, nf)).copy()
            self.covars_weight = np.broadcast_to(
                self.covars_weight, (nc, nm, nf)).copy()
        elif self.covariance_type == "spherical":
            self.covars_prior = np.broadcast_to(
                self.covars_prior, (nc, nm)).copy()
            self.covars_weight = np.broadcast_to(
                self.covars_weight, (nc, nm)).copy()

    def _check(self):
        super()._check()
        if not hasattr(self, "n_features"):
            self.n_features = self.means_.shape[2]
        nc = self.n_components
        nf = self.n_features
        nm = self.n_mix

        self._init_covar_priors()
        self._fix_priors_shape()

        # Checking covariance type
        if self.covariance_type not in COVARIANCE_TYPES:
            raise ValueError("covariance_type must be one of {}"
                             .format(COVARIANCE_TYPES))

        self.weights_ = np.array(self.weights_)
        # Checking mixture weights' shape
        if self.weights_.shape != (nc, nm):
            raise ValueError("mixture weights must have shape "
                             "(n_components, n_mix), actual shape: {}"
                             .format(self.weights_.shape))

        # Checking mixture weights' mathematical correctness
        if not np.allclose(np.sum(self.weights_, axis=1), np.ones(nc)):
            raise ValueError("mixture weights must sum up to 1")

        # Checking means' shape
        self.means_ = np.array(self.means_)
        if self.means_.shape != (nc, nm, nf):
            raise ValueError("mixture means must have shape "
                             "(n_components, n_mix, n_features), "
                             "actual shape: {}".format(self.means_.shape))

        # Checking covariances' shape
        self.covars_ = np.array(self.covars_)
        covars_shape = self.covars_.shape
        needed_shapes = {
            "spherical": (nc, nm),
            "tied": (nc, nf, nf),
            "diag": (nc, nm, nf),
            "full": (nc, nm, nf, nf),
        }
        needed_shape = needed_shapes[self.covariance_type]
        if covars_shape != needed_shape:
            raise ValueError("{!r} mixture covars must have shape {}, "
                             "actual shape: {}"
                             .format(self.covariance_type,
                                     needed_shape, covars_shape))

        # Checking covariances' mathematical correctness
        from scipy import linalg

        if (self.covariance_type == "spherical" or
                self.covariance_type == "diag"):
            if np.any(self.covars_ < 0):
                raise ValueError("{!r} mixture covars must be non-negative"
                                 .format(self.covariance_type))
            if np.any(self.covars_ == 0):
                _log.warning("Degenerate mixture covariance")
        elif self.covariance_type == "tied":
            for i, covar in enumerate(self.covars_):
                if not np.allclose(covar, covar.T):
                    raise ValueError("Covariance of state #{} is not symmetric"
                                     .format(i))
                min_eigvalsh = np.linalg.eigvalsh(covar).min()
                if min_eigvalsh < 0:
                    raise ValueError("Covariance of state #{} is not positive "
                                     "definite".format(i))
                if min_eigvalsh == 0:
                    _log.warning("Covariance of state #%d has a null "
                                 "eigenvalue.", i)
        elif self.covariance_type == "full":
            for i, mix_covars in enumerate(self.covars_):
                for j, covar in enumerate(mix_covars):
                    if not np.allclose(covar, covar.T):
                        raise ValueError(
                            "Covariance of state #{}, mixture #{} is not "
                            "symmetric".format(i, j))
                    min_eigvalsh = np.linalg.eigvalsh(covar).min()
                    if min_eigvalsh < 0:
                        raise ValueError(
                            "Covariance of state #{}, mixture #{} is not "
                            "positive definite".format(i, j))
                    if min_eigvalsh == 0:
                        _log.warning("Covariance of state #%d, mixture #%d "
                                     "has a null eigenvalue.", i, j)

    def _generate_sample_from_state(self, state, random_state=None):
        if random_state is None:
            random_state = self.random_state
        random_state = check_random_state(random_state)

        cur_weights = self.weights_[state]
        i_gauss = random_state.choice(self.n_mix, p=cur_weights)
        if self.covariance_type == 'tied':
            # self.covars_.shape == (n_components, n_features, n_features)
            # shouldn't that be (n_mix, ...)?
            covs = self.covars_
        else:
            covs = self.covars_[:, i_gauss]
            covs = fill_covars(covs, self.covariance_type,
                               self.n_components, self.n_features)
        return random_state.multivariate_normal(
            self.means_[state, i_gauss], covs[state]
        )

    def _compute_log_weighted_gaussian_densities(self, X, i_comp):
        cur_means = self.means_[i_comp]
        cur_covs = self.covars_[i_comp]
        if self.covariance_type == 'spherical':
            cur_covs = cur_covs[:, None]
        log_cur_weights = np.log(self.weights_[i_comp])

        return log_multivariate_normal_density(
            X, cur_means, cur_covs, self.covariance_type
        ) + log_cur_weights

    def _compute_log_likelihood(self, X):
        n_samples, _ = X.shape
        res = np.zeros((n_samples, self.n_components))

        for i in range(self.n_components):
            log_denses = self._compute_log_weighted_gaussian_densities(X, i)
            with np.errstate(under="ignore"):
                res[:, i] = logsumexp(log_denses, axis=1)

        return res

    def _initialize_sufficient_statistics(self):
        stats = super()._initialize_sufficient_statistics()
        stats['n_samples'] = 0
        stats['post_comp_mix'] = None
        stats['post_mix_sum'] = np.zeros((self.n_components, self.n_mix))
        stats['post_sum'] = np.zeros(self.n_components)
        stats['samples'] = None
        stats['centered'] = None
        return stats

    def _accumulate_sufficient_statistics(self, stats, X, framelogprob,
                                          post_comp, fwdlattice, bwdlattice):

        # TODO: support multiple frames

        super()._accumulate_sufficient_statistics(
            stats, X, framelogprob, post_comp, fwdlattice, bwdlattice
        )

        n_samples, _ = X.shape

        stats['n_samples'] = n_samples
        stats['samples'] = X

        post_mix = np.zeros((n_samples, self.n_components, self.n_mix))
        for p in range(self.n_components):
            log_denses = self._compute_log_weighted_gaussian_densities(X, p)
            log_normalize(log_denses, axis=-1)
            with np.errstate(under="ignore"):
                post_mix[:, p, :] = np.exp(log_denses)

        with np.errstate(under="ignore"):
            post_comp_mix = post_comp[:, :, None] * post_mix
        stats['post_comp_mix'] = post_comp_mix

        stats['post_mix_sum'] = np.sum(post_comp_mix, axis=0)
        stats['post_sum'] = np.sum(post_comp, axis=0)

        stats['centered'] = X[:, None, None, :] - self.means_

    def _do_mstep(self, stats):
        super()._do_mstep(stats)
        nc = self.n_components
        nf = self.n_features
        nm = self.n_mix

        n_samples = stats['n_samples']

        # Maximizing weights
        if 'w' in self.params:
            alphas_minus_one = self.weights_prior - 1
            new_weights_numer = stats['post_mix_sum'] + alphas_minus_one
            new_weights_denom = (
                stats['post_sum'] + np.sum(alphas_minus_one, axis=1)
            )[:, None]
            new_weights = new_weights_numer / new_weights_denom

            self.weights_ = new_weights

        # Maximizing means
        if 'm' in self.params:
            lambdas, mus = self.means_weight, self.means_prior
            new_means_numer = (
                np.einsum('ijk,il->jkl',
                          stats['post_comp_mix'], stats['samples'])
                + lambdas[:, :, None] * mus
            )
            new_means_denom = (stats['post_mix_sum'] + lambdas)[:, :, None]
            new_means = new_means_numer / new_means_denom

            self.means_ = new_means

        # Maximizing covariances
        if 'c' in self.params:
            centered_means = self.means_ - mus

        def outer_f(x):  # Outer product over features.
            return x[..., :, None] * x[..., None, :]

            if self.covariance_type == 'full':
                centered_dots = outer_f(stats['centered'])
                centered_means_dots = outer_f(centered_means)

                psis_t = np.transpose(self.covars_prior, axes=(0, 1, 3, 2))
                nus = self.covars_weight

                new_cov_numer = (
                    np.einsum('ijk,ijklm->jklm',
                              stats['post_comp_mix'], centered_dots)
                    + psis_t
                    + lambdas[:, :, None, None] * centered_means_dots
                )
                new_cov_denom = (
                    stats['post_mix_sum'] + 1 + nus + nf + 1)[:, :, None, None]
                new_cov = new_cov_numer / new_cov_denom

            elif self.covariance_type == 'diag':
                centered2 = stats['centered'] ** 2
                centered_means2 = centered_means ** 2

                alphas = self.covars_prior
                betas = self.covars_weight

                new_cov_numer = (
                    np.einsum('ijk,ijkl->jkl',
                              stats['post_comp_mix'], centered2)
                    + lambdas[:, :, None] * centered_means2
                    + 2 * betas
                )
                new_cov_denom = (
                    stats['post_mix_sum'][:, :, None] + 1 + 2 * (alphas + 1))
                new_cov = new_cov_numer / new_cov_denom

            elif self.covariance_type == 'spherical':
                centered_norm2 = np.sum(stats['centered'] ** 2, axis=-1)
                centered_means_norm2 = np.sum(centered_means ** 2, axis=-1)

                alphas = self.covars_prior
                betas = self.covars_weight

                new_cov_numer = (
                    np.einsum(
                        'ijk,ijk->jk', stats['post_comp_mix'], centered_norm2)
                    + lambdas * centered_means_norm2
                    + 2 * betas
                )
                new_cov_denom = (
                    nf * (stats['post_mix_sum'] + 1) + 2 * (alphas + 1))
                new_cov = new_cov_numer / new_cov_denom

            elif self.covariance_type == 'tied':
                centered_dots = outer_f(stats['centered'])
                centered_means_dots = outer_f(centered_means)

                psis_t = np.transpose(self.covars_prior, axes=(0, 2, 1))
                nus = self.covars_weight

                lambdas_cmdots_prod_sum = (
                    np.einsum('ij,ijkl->ikl', lambdas, centered_means_dots))

                new_cov_numer = (
                    np.einsum('ijk,ijklm->jlm',
                              stats['post_comp_mix'], centered_dots)
                    + lambdas_cmdots_prod_sum + psis_t)
                new_cov_denom = (
                    stats['post_sum'] + nm + nus + nf + 1)[:, None, None]
                new_cov = new_cov_numer / new_cov_denom

            self.covars_ = new_cov


import unittest
from datetime import datetime

import pandas as pd
from airflow import DAG

from dsbox.operators.data_operator import DataOperator
from dsbox.operators.data_unit import DataInputFileUnit, DataOutputFileUnit
from dsbox.utils import execute_dag


def drop_na_dataframe(dataframe, columns):
    dataframe = dataframe.dropna(subset=columns)
    return dataframe


class TestDagFile(unittest.TestCase):
    def test_building_two_operators_with_execution(self):
        # given
        dag = DAG(dag_id='test_dag_file', start_date=datetime.now())

        input_csv_unit = DataInputFileUnit('data/X.csv', sep=';')
        output_parquet_unit = DataOutputFileUnit('data/X_parsed.parquet', pandas_write_function_name='to_parquet')
        task_1 = DataOperator(operation_function=drop_na_dataframe,
                              params={'columns': ['ANNEEREALISATIONDIAGNOSTIC']},
                              input_unit=input_csv_unit,
                              output_unit=output_parquet_unit,
                              dag=dag, task_id='data_operator_csv_to_parquet')

        input_parquet_unit = DataInputFileUnit('data/X_parsed.parquet', pandas_read_function_name='read_parquet')
        output_csv_unit = DataOutputFileUnit('data/X_parsed_2.csv', index=False)
        task_2 = DataOperator(operation_function=drop_na_dataframe,
                              params={'columns': ['ANNEETRAVAUXPRECONISESDIAG']},
                              input_unit=input_parquet_unit,
                              output_unit=output_csv_unit,
                              dag=dag, task_id='data_operator_parquet_to_csv')

        task_2.set_upstream(task_1)

        # when
        execute_dag(dag, verbose=True)

        # then
        df = pd.read_csv('data/X_parsed_2.csv')
        self.assertEqual((7241, 27), df.shape)



import io

import pandas as pd
import psycopg2 as pg
import sqlalchemy


class DBconnector():
    """
    Simple DB connector to write pandas dataframe to RDBMS table or to read from a table to a pandas dataframe.
    """

    def __init__(self, username, password, hostname, port, dbname, baseprotocol=''):
        self.username = username
        self.password = password
        self.hostname = hostname
        self.port = port
        self.dbname = dbname
        self.basetype = baseprotocol

        self.uri = baseprotocol + self.username + ':' + self.password + '@' + self.hostname + '/' + self.dbname
        self.engine = sqlalchemy.create_engine(self.uri)

    def df_to_table(self, df, table_name, **kwargs):
        """
        Write pandas dataframe to table.

        Parameters
        ----------
        df: pandas.DataFrame
            dataframe to export
        table_name: str
            target table name
        """
        df.to_sql(table_name, self.engine, **kwargs)

    def table_to_df(self, table_name, **kwargs):
        """
        Returns pandas dataframe from table.

        Parameters
        ----------
        table_name: str
            Table to read

        Returns
        -------
        pandas.DataFrame
            Result dataframe.

        """
        return pd.read_sql_query('SELECT * from ' + table_name, self.engine, **kwargs)

    def check_table(self, table_name, schema=None):
        """
        Check if a table exists.

        Parameters
        ----------
        table_name: str
            table name to check
        schema: str, default=None
            schema where lays the table to check

        Returns
        -------
        bool:
             Returns True if table exists, else returns False.
        """
        if schema is None:
            return self.engine.has_table(table_name)
        else:
            return self.engine.has_table(table_name, schema=schema)


class DBconnectorPG(DBconnector):
    """
        Specialized Postgres DB connector to write pandas dataframe to RDBMS table or to read from a table to a pandas dataframe.
        Using bulk operations.
    """

    def __init__(self, username, password, hostname, port, dbname):
        super(DBconnectorPG, self).__init__(username, password, hostname, port, dbname, baseprotocol='postgres://')

        self.con = pg.connect(user=self.username, password=self.password,
                              host=self.hostname, port=self.port, dbname=self.dbname)
        self.con.autocommit = True

    def bulk_to_pg(self, df, table_name, to_pg_drop=False):
        """
        Writes a dataframe into a PG table with bulk operations.

        Parameters
        ----------
        df: pandas.DataFrame
            dataframe to write
        table_name: str
            target table name
        to_pg_drop: bool, default=False
            specify if target table must be overwritten
        """

        data = io.StringIO()
        df.to_csv(data, header=True, index=False, sep=',', quotechar='"')
        data.seek(0)
        cursor = self.con.cursor()
        if to_pg_drop:
            cursor.execute("DROP TABLE IF EXISTS " + table_name)
            empty_table = pd.io.sql.get_schema(df, table_name, con=self.engine)
            empty_table = empty_table.replace('"', '')
            cursor.execute(empty_table)
        sql_query = "COPY " + table_name + " FROM STDIN WITH DELIMITER ',' CSV HEADER QUOTE '" + '"' + "'"
        cursor.copy_expert(sql_query, data)
        cursor.connection.commit()

    def bulk_from_pg(self, table_name):
        """
        Read a PG table to dataframe with bulk operations.

        Parameters
        ----------
        table_name: str
            table to read

        Returns
        -------
        df: pandas.DataFrame
            dataframe result

        """
        buf = io.StringIO()
        cursor = self.con.cursor()
        sql_query = "COPY " + table_name + " TO STDOUT WITH DELIMITER ',' CSV HEADER QUOTE '" + '"' + "' FORCE QUOTE *"
        cursor.copy_expert(sql_query, buf)
        buf.seek(0)
        df = pd.read_csv(buf)
        buf.close()

        return df

    def check_table_pg(self, schema_name, table_name):
        """
        Check if a PG table exists.

        Parameters
        ----------
        schema_name: str
            schema where lays the table to check

        table_name: str
            table name to check

        Returns
        -------
        bool:
             Returns True if table exists, else returns False.
        """
        query = "SELECT 1" \
                " FROM information_schema.tables" \
                " WHERE table_schema = '" + schema_name + "'" + \
                " AND table_name = '" + table_name + "'"

        df = pd.read_sql_query(query, self.engine)
        if len(df) > 0:
            return True
        else:
            return False



import unittest
from datetime import datetime
import pandas as pd

from airflow import DAG

from dsbox.operators.data_operator import DataOperator
from dsbox.operators.data_unit import DataInputFileUnit, DataOutputFileUnit, DataOutputPlasmaUnit, DataInputPlasmaUnit
from dsbox.dbconnection.plasma_connector import PlasmaConnector
from dsbox.utils import execute_dag

from tests.config import socket_name, object_id


def drop_na_dataframe(dataframe, columns):
    dataframe = dataframe.dropna(subset=columns)
    return dataframe


class TestDagFile(unittest.TestCase):
    def test_building_two_operators_with_execution(self):
        # given
        plasma_connector = PlasmaConnector(socket_name)

        dag = DAG(dag_id='test_dag_plasma', start_date=datetime.now())

        input_csv_unit = DataInputFileUnit('data/X.csv', sep=';')
        output_plasma_unit = DataOutputPlasmaUnit(plasma_connector, object_id)
        task_1 = DataOperator(operation_function=drop_na_dataframe,
                              params={'columns': ['ANNEEREALISATIONDIAGNOSTIC']},
                              input_unit=input_csv_unit,
                              output_unit=output_plasma_unit,
                              dag=dag, task_id='data_operator_csv_to_plasma')

        input_plasma_unit = DataInputPlasmaUnit(plasma_connector, object_id)
        output_csv_unit = DataOutputFileUnit('data/X_parsed_22.csv', index=False)
        task_2 = DataOperator(operation_function=drop_na_dataframe,
                              params={'columns': ['ANNEETRAVAUXPRECONISESDIAG']},
                              input_unit=input_plasma_unit,
                              output_unit=output_csv_unit,
                              dag=dag, task_id='data_operator_plasma_to_csv')

        task_2.set_upstream(task_1)

        # when
        execute_dag(dag, verbose=True)

        # then
        df = pd.read_csv('data/X_parsed_22.csv')
        self.assertEqual((7241, 27), df.shape)





# Copyright 2013-2014 Lars Buitinck / University of Amsterdam
#
# Parts taken from scikit-learn, written by
#          Olivier Grisel
#          Gael Varoquaux
#          Andreas Mueller
#          Lars Buitinck
#          Alexandre Gramfort

import numpy as np
import scipy.sparse as sp

# XXX These are private helper functions from scikit-learn. We should copy
# the code over instead of importing them.
from sklearn.utils import check_random_state
from sklearn.utils.extmath import safe_sparse_dot

from .ctrans import count_trans
from .safeadd import safe_add
from .transmatrix import make_trans_matrix


def _assert_all_finite(X):
    X = np.asanyarray(X)
    # First try an O(n) time, O(1) space solution for the common case that
    # there everything is finite; fall back to O(n) space np.isfinite to
    # prevent false positives from overflow in sum method.
    if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
            and not np.isfinite(X).all()):
        raise ValueError("Input contains NaN, infinity"
                         " or a value too large for %r." % X.dtype)


def assert_all_finite(X):
    """Throw a ValueError if X contains NaN or infinity.

    Input MUST be an np.ndarray instance or a scipy.sparse matrix."""

    _assert_all_finite(X.data if sp.issparse(X) else X)


def array2d(X, dtype=None, order=None, copy=False):
    """Returns at least 2-d array with data from X"""
    if sp.issparse(X):
        raise TypeError('A sparse matrix was passed, but dense data '
                        'is required. Use X.toarray() to convert to dense.')
    X_2d = np.asarray(np.atleast_2d(X), dtype=dtype, order=order)
    _assert_all_finite(X_2d)
    if X is X_2d and copy:
        X_2d = np.copy(X_2d)
    return X_2d


def _atleast2d_or_sparse(X, dtype, order, copy, sparse_class, convmethod,
                         check_same_type):
    if sp.issparse(X):
        if check_same_type(X) and X.dtype == dtype:
            X = getattr(X, convmethod)(copy=copy)
        elif dtype is None or X.dtype == dtype:
            X = getattr(X, convmethod)()
        else:
            X = sparse_class(X, dtype=dtype)
        _assert_all_finite(X.data)
        X.data = np.array(X.data, copy=False, order=order)
    else:
        X = array2d(X, dtype=dtype, order=order, copy=copy)
    return X


def atleast2d_or_csr(X, dtype=None, order=None, copy=False):
    """Like numpy.atleast_2d, but converts sparse matrices to CSR format

    Also, converts np.matrix to np.ndarray.
    """
    return _atleast2d_or_sparse(X, dtype, order, copy, sp.csr_matrix,
                                "tocsr", sp.isspmatrix_csr)


def validate_lengths(n_samples, lengths):
    """Validate lengths array against n_samples.

    Parameters
    ----------
    n_samples : integer
        Total number of samples.

    lengths : array-like of integers, shape (n_sequences,), optional
        Lengths of individual sequences in the input.

    Returns
    -------
    start : array of integers, shape (n_sequences,)
        Start indices of sequences.

    end : array of integers, shape (n_sequences,)
        One-past-the-end indices of sequences.
    """
    if lengths is None:
        lengths = [n_samples]
    lengths = np.asarray(lengths, dtype=np.int32)
    if lengths.sum() > n_samples:
        msg = "More than {0:d} samples in lengths array {1!s}"
        raise ValueError(msg.format(n_samples, lengths))

    end = np.cumsum(lengths)
    start = end - lengths

    return start, end



import unittest
import numpy as np

from dsbox.ml.markov.hmmlearn import normalize, fill_covars


class TestUtils(unittest.TestCase):
    def test_normalize(self):
        A = np.random.normal(42., size=128)
        A[np.random.choice(len(A), size=16)] = 0.0
        self.assertTrue((A == 0.0).any())
        normalize(A)
        self.assertTrue(np.allclose(A.sum(), 1.))

    def test_normalize_along_axis(self):
        A = np.random.normal(42., size=(128, 4))
        for axis in range(A.ndim):
            A[np.random.choice(len(A), size=16), axis] = 0.0
            self.assertTrue((A[:, axis] == 0.0).any())
            normalize(A, axis=axis)
            self.assertTrue(np.allclose(A.sum(axis=axis), 1.))

    def test_fill_covars(self):
        full = np.arange(12).reshape(3, 2, 2) + 1
        np.testing.assert_equal(fill_covars(full, 'full', 3, 2), full)

        diag = np.arange(6).reshape(3, 2) + 1
        expected = np.array([[[1, 0], [0, 2]],
                             [[3, 0], [0, 4]],
                             [[5, 0], [0, 6]]])
        np.testing.assert_equal(fill_covars(diag, 'diag', 3, 2), expected)

        tied = np.arange(4).reshape(2, 2) + 1
        expected = np.array([[[1, 2], [3, 4]],
                             [[1, 2], [3, 4]],
                             [[1, 2], [3, 4]]])
        np.testing.assert_equal(fill_covars(tied, 'tied', 3, 2), expected)

        spherical = np.array([1, 2, 3])
        expected = np.array([[[1, 0], [0, 1]],
                             [[2, 0], [0, 2]],
                             [[3, 0], [0, 3]]])
        np.testing.assert_equal(
            fill_covars(spherical, 'spherical', 3, 2), expected)



import unittest

from numpy.testing import assert_array_equal

import numpy as np
from scipy.sparse import coo_matrix, csc_matrix
from sklearn.base import clone

from dsbox.ml.markov.seqlearn import StructuredPerceptron


class TestSeqPerceptron(unittest.TestCase):
    def test_perceptron(self):
        X = [[0, 1, 0],
             [0, 1, 0],
             [1, 0, 0],
             [0, 1, 0],
             [1, 0, 0],
             [0, 0, 1],
             [0, 0, 1],
             [0, 1, 0],
             [1, 0, 0],
             [1, 0, 0]]

        y = [0, 0, 0, 0, 0, 1, 1, 0, 2, 2]

        clf = StructuredPerceptron(verbose=False, random_state=37, max_iter=15)
        clf.fit(X, y, [len(y)])
        assert_array_equal(y, clf.predict(X))

        # Try again with string labels and sparse input.
        y_str = np.array(["eggs", "ham", "spam"])[y]

        clf = clone(clf)
        clf.fit(csc_matrix(X), y_str, [len(y_str)])
        assert_array_equal(y_str, clf.predict(coo_matrix(X)))

        X2 = np.vstack([X, X])
        y2 = np.hstack([y_str, y_str])
        assert_array_equal(y2, clf.predict(X2, lengths=[len(y), len(y)]))

    def test_perceptron_single_iter(self):
        """Assert that averaging works after a single iteration."""
        clf = StructuredPerceptron(max_iter=1)
        self.assertIsNotNone(clf.fit([[1, 2, 3]], [1], [1]))



import unittest

import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN, KMeans
from sklearn.covariance import EmpiricalCovariance, MinCovDet
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.mixture import GaussianMixture

from dsbox.ml.outliers import CovarianceOutliers, GaussianProcessOutliers
from dsbox.ml.outliers import GMMOutliers, ClusteringOutliers
from dsbox.ml.outliers import KMeansOneClusterOutliers, KMeansIterativeOneClusterOutliers
from dsbox.ml.outliers import MADOutliers, FFTOutliers


class CovarianceOutliersTest(unittest.TestCase):
    def test_covarianceoutliers_constructor_should_accept_different_scikit_covariance_estimators(self):
        # given
        robust_cov = MinCovDet()
        emp_cov = EmpiricalCovariance()

        # when
        cov_outliers_1 = CovarianceOutliers(emp_cov)
        cov_outliers_2 = CovarianceOutliers(robust_cov)

        # then
        self.assertTrue(isinstance(cov_outliers_1, CovarianceOutliers))
        self.assertTrue(isinstance(cov_outliers_2, CovarianceOutliers))

    def test_covarianceoutliers_predict_proba_gives_biggest_proba_to_biggest_outlier(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        cov_outliers = CovarianceOutliers()
        cov_outliers.fit(df)
        probas = cov_outliers.predict_proba(df)
        outlier_index = np.argmax(probas)

        # then
        outlier_index_true = 6
        self.assertEqual(outlier_index_true, outlier_index)

    def test_covarianceoutliers_predict_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        cov_outliers = CovarianceOutliers()
        cov_outliers.fit(df)
        outliers = cov_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, False, False, True, True, False, False, False,
                         False, False, False]
        self.assertListEqual(outliers_true, outliers.tolist())


class GaussianProcessOutliersTest(unittest.TestCase):
    def test_gpoutliers_predict_should_return_correct_values(self):
        # given
        data = np.random.random_sample(1000) * 2 - 1
        data[300] = 5
        data[700] = -6
        df = pd.DataFrame(data)

        # when
        gp_outliers = GaussianProcessOutliers(GaussianProcessRegressor(alpha=0.9, normalize_y=True), n_samples=100)
        gp_outliers.fit(df)
        outliers = gp_outliers.predict(df, confidence=0.999)

        # then
        outlier_positions_true = [300, 700]
        self.assertTrue(outliers[outlier_positions_true[0]])
        self.assertTrue(outliers[outlier_positions_true[1]])


class KMeansOneClusterOutliersTest(unittest.TestCase):
    def test_kmeansonecluster_outliers_predict_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        kmoc_outliers = KMeansOneClusterOutliers()
        kmoc_outliers.fit(df)
        outliers = kmoc_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, False, False, True, True, False, False, False,
                         False, False, False]
        self.assertListEqual(outliers_true, outliers.tolist())


class KMeansIterativeOneClusterOutliersTest(unittest.TestCase):
    def test_kmeans_iterative_onecluster_outliers_predict_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        kmoc_outliers = KMeansIterativeOneClusterOutliers()
        outliers = kmoc_outliers.fit_predict(df)

        # then
        outliers_true = [2., 1., 1., 2., 2., -1., 0., 0., 1., 1., -1., 1., 1., 2.]
        self.assertListEqual(outliers_true, outliers.tolist())


class GMMOutliersTest(unittest.TestCase):
    def test_gmm_outliers_predict_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        gmm_outliers = GMMOutliers()
        gmm_outliers.fit(df)
        outliers = gmm_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, False, False, True, True, False, False, False,
                         False, False, False]
        self.assertListEqual(outliers_true, outliers.tolist())


class ClusteringOutliersTest(unittest.TestCase):
    def test_clustering_outliers_predict_proba_with_unclustered_strategy_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=DBSCAN(min_samples=2), strategy='unclustered')
        clustering_outliers.fit(df)
        outliers = clustering_outliers.predict_proba(df)

        # then
        outliers_true = [0., 0., 0., 0., 1., 0., 1., 1., 1., 1., 0., 0.,
                         1., 0.]
        self.assertListEqual(outliers_true, outliers.tolist())

    def test_clustering_outliers_predict_with_unclustered_strategy_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=DBSCAN(min_samples=2), strategy='unclustered')
        clustering_outliers.fit(df)
        outliers = clustering_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, True, False, True, True, True, True, False, False,
                         True, False]
        self.assertListEqual(outliers_true, outliers.tolist())

    def test_clustering_outliers_predict_proba_with_frontier_strategy_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=KMeans(n_clusters=2), strategy='frontier')
        clustering_outliers.fit(df)
        outliers_probas = clustering_outliers.predict_proba(df)

        # then
        outliers_true = [0.01861993, 0.02190581, 0.02190581, 0.01861993, 0.0109529, 0.01533406
            , 0.00196078, 0.00196078, 0.08324206, 0.03066813, 0.01533406, 0.02190581
            , 0.02738226, 0.01861993]
        self.assertListEqual(outliers_true, np.round(outliers_probas, 8).tolist())

    def test_clustering_outliers_predict_with_frontier_strategy_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=KMeans(n_clusters=2), strategy='frontier')
        clustering_outliers.fit(df)
        outliers = clustering_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, False, False, False, False, True, True, False, False,
                         True, False]
        self.assertListEqual(outliers_true, outliers.tolist())

    def test_clustering_outliers_predict_proba_with_size_strategy_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=KMeans(n_clusters=2), strategy='size')
        clustering_outliers.fit(df)
        outliers_probas = clustering_outliers.predict_proba(df)

        # then
        outliers_true = [0.14285714, 0.14285714, 0.14285714, 0.14285714, 0.14285714, 0.14285714
            , 0.85714286, 0.85714286, 0.14285714, 0.14285714, 0.14285714, 0.14285714
            , 0.14285714, 0.14285714]
        self.assertListEqual(outliers_true, np.round(outliers_probas, 8).tolist())

    def test_clustering_outliers_predict_with_size_strategy_and_kmeans_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=KMeans(n_clusters=2), strategy='size')
        clustering_outliers.fit(df)
        outliers = clustering_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, False, False, True, True, False, False, False,
                         False, False, False]
        self.assertListEqual(outliers_true, outliers.tolist())

    def test_clustering_outliers_predict_with_size_strategy_and_gmm_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=GaussianMixture(n_components=2), strategy='size')
        clustering_outliers.fit(df)
        outliers = clustering_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, False, False, True, True, False, False, False,
                         False, False, False]
        self.assertListEqual(outliers_true, outliers.tolist())

    def test_clustering_outliers_predict_with_size_strategy_and_dbscan_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=DBSCAN(min_samples=2), strategy='size',
                                                 threshold=0.8)
        clustering_outliers.fit(df)
        outliers = clustering_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, False, True, False, False, False, False, True, False
            , False, False]
        self.assertListEqual(outliers_true, outliers.tolist())

    def test_clustering_outliers_predict_proba_with_density_strategy_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=KMeans(n_clusters=2), strategy='density')
        clustering_outliers.fit(df)
        outliers_probas = clustering_outliers.predict_proba(df)

        # then
        outliers_true = [0.26737475, 0.26737475, 0.26737475, 0.26737475, 0.26737475, 0.26737475
            , 0.73262525, 0.73262525, 0.26737475, 0.26737475, 0.26737475, 0.26737475
            , 0.26737475, 0.26737475]
        self.assertListEqual(outliers_true, np.round(outliers_probas, 8).tolist())

    def test_clustering_outliers_predict_with_density_strategy_and_dbscan_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        clustering_outliers = ClusteringOutliers(cluster_estimator=DBSCAN(min_samples=2), strategy='density',
                                                 threshold=0.8)
        clustering_outliers.fit(df)
        outliers = clustering_outliers.predict(df)

        # then
        outliers_true = [False, False, False, False, True, False, True, True, True, True, False, False
            , True, False]
        self.assertListEqual(outliers_true, outliers.tolist())


class WrappersTest(unittest.TestCase):
    def test_mad_outliers_predict_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        mad_outliers = MADOutliers()
        outliers = mad_outliers.fit_predict(df)

        # then
        outliers_true = [False, False, False, False,  True, False,  True,  True,  True,
           False, False, False,  True, False]

        self.assertListEqual(outliers_true, outliers.tolist())

    def test_fft_outliers_predict_should_return_correct_values(self):
        # given
        df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        fft_outliers = FFTOutliers(threshold=0.9)
        outliers = fft_outliers.fit_predict(df)

        # then
        outliers_true = [False, False, False, False, False, False, False, True, False, False, False, False
            , False, False]

        self.assertListEqual(outliers_true, outliers.tolist())


if __name__ == '__main__':
    unittest.main()



import unittest

from dsbox.utils import FilenameGenerator


class TestFilenameGenerator(unittest.TestCase):
    def test_generate_filename(self):

        # given
        fg = FilenameGenerator()

        # when
        filename_10 = fg.generate_filename(pos=10)
        filename_9 = fg.generate_filename(pos=9)

        # then
        expected_filename_10 = 'X_1a92'
        expected_filename_9 = 'X_ae7d'

        self.assertEquals(expected_filename_10, filename_10)
        self.assertEquals(expected_filename_9, filename_9)

if __name__ == '__main__':
    unittest.main()



import numpy as np
from dsbox.ml.neural_networks import KerasFactory

from tensorflow.keras import backend as K
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Input, Dense, Conv2D, MaxPooling2D, UpSampling2D, Lambda
from tensorflow.keras.losses import mse, binary_crossentropy
from tensorflow.keras.models import Model
from sklearn.base import BaseEstimator, ClassifierMixin

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"

__source__ = "https://blog.keras.io/building-autoencoders-in-keras.html"


class SimpleAutoEncoderFactory(KerasFactory):
    def create_model(self, encoding_dim=32, input_size=784):
        """
        Build a simple keras model autoencoder with 3 dense layers.
        
        Mostly used for data denoising or outliers detection.
        
        Parameters
        ----------
        encoding_dim: int, optional (default=32)
            intermediary layer size
            
        input_size: int, optional (default=784)
            original input data size

        Returns
        -------
        3 Keras models (autoencoder, encoder, decoder)
        """
        input_layer = Input(shape=(input_size,))
        encoded_layer = Dense(encoding_dim, activation='relu')(input_layer)
        decoded_layer = Dense(input_size, activation='sigmoid')(encoded_layer)

        autoencoder = Model(input_layer, decoded_layer)
        encoder = Model(input_layer, encoded_layer)

        encoded_input = Input(shape=(encoding_dim,))
        decoder_layer = autoencoder.layers[-1]
        decoder = Model(encoded_input, decoder_layer(encoded_input))

        autoencoder.compile(optimizer='adadelta', loss='binary_crossentropy')

        return autoencoder, encoder, decoder


class SparseAutoEncoderFactory(KerasFactory):
    def create_model(self, encoding_dim=32, input_size=784):
        """
        Build a sparse keras model autoencoder with 3 dense layers.
        It adds a regularization constraint in the intermediary layer.
        
        Mostly used for data denoising or outliers detection.
        
        Parameters
        ----------
        encoding_dim: int, optional (default=32)
           intermediary layer size
        
        input_size: int, optional (default=784)
           original input data size
        
        Returns
        -------
        3 Keras models (autoencoder, encoder, decoder)
        """
        input_layer = Input(shape=(input_size,))
        encoded_layer = Dense(encoding_dim, activation='relu',
                              activity_regularizer=regularizers.l1(10e-5))(input_layer)
        decoded_layer = Dense(input_size, activation='sigmoid')(encoded_layer)

        autoencoder = Model(input_layer, decoded_layer)
        encoder = Model(input_layer, encoded_layer)

        encoded_input = Input(shape=(encoding_dim,))
        decoder_layer = autoencoder.layers[-1]
        decoder = Model(encoded_input, decoder_layer(encoded_input))

        autoencoder.compile(optimizer='adadelta', loss='binary_crossentropy')

        return autoencoder, encoder, decoder


class DeepAutoEncoderFactory(KerasFactory):
    def create_model(self, input_size=784, start_layer_size=128, sub_layers_level=3):
        """
        Build a deep keras model autoencoder with x dense layers. For each sub-layer,
        the start_layer size is divided by 2.
        With default parameters (start_layer_size=128, sub_layers_level=3), it builds an 
        architecture like this:
        input_size-128-64-32-64-128-input_size
        
        Mostly used for data denoising or outliers detection.
        
        Parameters
        ---------- 
        input_size: int, optional (default=784)
           original input data size
           
        start_layer_size: int, optional (default=128)
            size of the first layer after the input one
            
        sub_layers_level: int, optional (default=3)
            amount of sub-levels to build
        
        Returns
        -------
        2 Keras models (autoencoder, encoder)
        """
        input_layer = Input(shape=(input_size,))
        encoded_layer = Dense(start_layer_size, activation='relu')(input_layer)
        next_layer_size = int(start_layer_size)
        for i in range(1, sub_layers_level):
            next_layer_size = int(next_layer_size / 2)
            encoded_layer = Dense(next_layer_size, activation='relu')(encoded_layer)

        next_layer_size = next_layer_size * 2
        decoded_layer = Dense(next_layer_size, activation='relu')(encoded_layer)

        for i in range(2, sub_layers_level):
            next_layer_size = next_layer_size * 2
            decoded_layer = Dense(next_layer_size, activation='relu')(decoded_layer)

        decoded_layer = Dense(input_size, activation='sigmoid')(decoded_layer)

        autoencoder = Model(input_layer, decoded_layer)
        encoder = Model(input_layer, encoded_layer)

        autoencoder.compile(optimizer='adadelta', loss='binary_crossentropy')

        return autoencoder, encoder


class CNNAutoEncoder2DFactory(KerasFactory):
    def create_model(self, shape=(28, 28, 1)):
        """
        Build a 2D convolutional keras model autoencoder.

        Mostly used for image denoising or image outliers detection.

        Parameters
        ----------           
        shape: int, optional (default=(28,28,1))
           original input data size

        Returns
        -------
        2 Keras models (autoencoder, encoder)
        """
        layer_size = np.int(np.round((np.sqrt(shape[0] * shape[1]) * 0.1)))
        pool_size = np.int(np.round(layer_size * 2 / 3.0))

        input_img = Input(shape=shape)

        x = Conv2D(16, (layer_size, layer_size), activation='relu', padding='same')(input_img)
        x = MaxPooling2D((pool_size, pool_size), padding='same')(x)
        x = Conv2D(8, (layer_size, layer_size), activation='relu', padding='same')(x)
        x = MaxPooling2D((pool_size, pool_size), padding='same')(x)
        x = Conv2D(8, (layer_size, layer_size), activation='relu', padding='same')(x)
        encoded_layer = MaxPooling2D((pool_size, pool_size), padding='same')(x)

        x = Conv2D(8, (layer_size, layer_size), activation='relu', padding='same')(encoded_layer)
        x = UpSampling2D((pool_size, pool_size))(x)
        x = Conv2D(8, (layer_size, layer_size), activation='relu', padding='same')(x)
        x = UpSampling2D((pool_size, pool_size))(x)
        x = Conv2D(16, (layer_size, layer_size), activation='relu')(x)
        x = UpSampling2D((pool_size, pool_size))(x)
        decoded_layer = Conv2D(1, (layer_size, layer_size), activation='sigmoid', padding='same')(x)

        autoencoder = Model(input_img, decoded_layer)
        encoder = Model(input_img, encoded_layer)

        autoencoder.compile(optimizer='adadelta', loss='binary_crossentropy')

        return autoencoder, encoder


class CNNDenoisingAutoEncoder2DFactory(KerasFactory):
    def create_model(self, shape=(28, 28, 1)):
        """
        Build a 2D convolutional keras model autoencoder specialized in image denoising.

        Parameters
        ----------           
        shape: int, optional (default=(28,28,1))
           original input data size

        Returns
        -------
        2 Keras models (autoencoder, encoder)
        """
        layer_size = np.int(np.round((np.sqrt(shape[0] * shape[1]) * 0.1)))
        pool_size = np.int(np.round(layer_size * 2 / 3.0))

        input_img = Input(shape=shape)

        x = Conv2D(32, (layer_size, layer_size), activation='relu', padding='same')(input_img)
        x = MaxPooling2D((pool_size, pool_size), padding='same')(x)
        x = Conv2D(32, (layer_size, layer_size), activation='relu', padding='same')(x)
        encoded_layer = MaxPooling2D((pool_size, pool_size), padding='same')(x)

        x = Conv2D(32, (layer_size, layer_size), activation='relu', padding='same')(encoded_layer)
        x = UpSampling2D((pool_size, pool_size))(x)
        x = Conv2D(32, (layer_size, layer_size), activation='relu', padding='same')(x)
        x = UpSampling2D((pool_size, pool_size))(x)
        decoded_layer = Conv2D(1, (layer_size, layer_size), activation='sigmoid', padding='same')(x)

        autoencoder = Model(input_img, decoded_layer)
        encoder = Model(input_img, encoded_layer)

        autoencoder.compile(optimizer='adadelta', loss='binary_crossentropy')

        return autoencoder, encoder


class VariationalAutoEncoderFactory(KerasFactory):
    def create_model(self, input_size=784, intermediate_dim=256, latent_dim=16, epsilon_std=1.0, vae_loss='xent'):
        """
        Build a keras model variational autoencoder.
        
        First, an encoder network turns the input samples x into two parameters in a latent space. Then, we randomly 
        sample similar points z from the latent normal distribution that is assumed to generate the data, 
        via z = z_mean + exp(z_log_sigma) * epsilon, where epsilon is a random normal tensor. Finally, a decoder 
        network maps these latent space points back to the original input data.

        The parameters of the model are trained via two loss functions: a reconstruction loss forcing the decoded 
        samples to match the initial inputs (just like in our previous autoencoders), and the KL divergence between the
        learned latent distribution and the prior distribution, acting as a regularization term. You could actually get 
        rid of this latter term entirely, although it does help in learning well-formed latent spaces and reducing 
        overfitting to the training data.
        
        Parameters
        ----------  
        input_size: int, optional (default=784)
           original input data size
           
        intermediate_dim: int, optional (default=16)
           size of the dense layers, placed after the input layer and before the output layer
           
        latent_dim: int, optional (default=16)
           size of the intermediary layer
           
        epsilon_std, float, optional (default=1.0)
            std value used for the normal distribution used to generate data into the latent space
            
        vae_loss: str, {'mse', 'xent'} optional, default 'xent'
            loss function used for reconstruction, MSE or binary Cross Entropy
        
        Returns
        -------
        3 Keras models (autoencoder, encoder, decoder)
        """
        self.latent_dim = latent_dim
        self.epsilon_std = epsilon_std

        input_shape = (input_size,)
        inputs = Input(shape=input_shape, name='encoder_input')
        x = Dense(intermediate_dim, activation='relu')(inputs)
        z_mean = Dense(latent_dim, name='z_mean')(x)
        z_log_var = Dense(latent_dim, name='z_log_var')(x)

        z = Lambda(self.sampling, output_shape=(latent_dim,), name='z')([z_mean, z_log_var])

        encoder = Model(inputs, [z_mean, z_log_var, z], name='encoder')

        latent_inputs = Input(shape=(latent_dim,), name='z_sampling')
        x = Dense(intermediate_dim, activation='relu')(latent_inputs)
        outputs = Dense(input_size, activation='sigmoid')(x)

        decoder = Model(latent_inputs, outputs, name='decoder')

        outputs = decoder(encoder(inputs)[2])
        vae = Model(inputs, outputs, name='vae_mlp')

        reconstruction_loss = None
        if vae_loss == 'mse':
            reconstruction_loss = mse(inputs, outputs)
        if vae_loss == 'xent':
            reconstruction_loss = binary_crossentropy(inputs, outputs)

        if reconstruction_loss is None:
            raise (ValueError("vae_loss parameter should take a valid value, 'mse' or 'xent'."))

        reconstruction_loss *= input_size
        kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
        kl_loss = K.sum(kl_loss, axis=-1)
        kl_loss *= -0.5
        vae_loss = K.mean(reconstruction_loss + kl_loss)
        vae.add_loss(vae_loss)
        vae.compile(optimizer='adam')

        return vae, encoder, decoder

    def sampling(self, args):
        z_mean, z_log_var = args
        epsilon = K.random_normal(shape=(K.shape(z_mean)[0], self.latent_dim), mean=0., stddev=self.epsilon_std)
        return z_mean + K.exp(z_log_var / 2) * epsilon


class AutoEncoderClassifier(BaseEstimator, ClassifierMixin):
    """
    Wrapper used to be scikit compliant, as our autoencoder factories returns more than one model.
    """

    def __init__(self, keras_autoencoder, keras_encoder):
        self.autoencoder = keras_autoencoder
        self.encoder = keras_encoder

    def fit(self, X, y, **kwargs):
        return self.autoencoder.fit(X, y, **kwargs)

    def predict(self, X):
        self.encoded_data_ = self.encoder.predict(X)
        return self.autoencoder.predict(X)



from abc import ABCMeta, abstractmethod

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class KerasFactory:
    """
    Abstract class template for all keras neural nets factories
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def create_model(self, **kwargs):
        pass



import unittest

import numpy as np

from dsbox.ml.outliers import median_absolute_deviation, mad_outliers, fft_outliers, double_median_absolute_deviation


class UtilsOutliersTest(unittest.TestCase):
    def test_mad(self):
        # given
        a = [1, 2, 3, 3, 4, 4, 4, 5, 5.5, 6, 6, 6.5, 7, 7, 7.5, 8, 9, 12, 52, 90]

        # when
        mad_value = median_absolute_deviation(a)

        # then
        self.assertEqual(2.0, mad_value)

    def test_double_mad(self):
        # given
        a = [1, 4, 4, 4, 5, 5, 5, 5, 7, 7, 8, 10, 16, 30]

        # when
        double_mad_values = double_median_absolute_deviation(a)

        # then
        self.assertEqual((1, 11), double_mad_values)

    def test_mad_outliers(self):
        # given
        a = np.array([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        outliers = mad_outliers(a, z_score_coeff=1.0)

        # then
        y_true = [False, False, False, False, True, False, True, True, True, True, False, False, True, False]

        self.assertListEqual(y_true, outliers.tolist())

    def test_fft_outliers(self):
        # given
        a = np.array([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])

        # when
        outliers = fft_outliers(a, freq_cut_index=12, outlier_proportion=0.2)

        # then
        y_true = [False, False, False, False, False, False, True, True, False, False, False, False, False, False]

        self.assertListEqual(y_true, outliers.tolist())


if __name__ == '__main__':
    unittest.main()



from . import _utils
from .stats import log_multivariate_normal_density
from .base import _BaseHMM
from .utils import iter_from_X_lengths, normalize, fill_covars
from .hmm import GMMHMM, GaussianHMM, MultinomialHMM

__all__ = ["GMMHMM", "GaussianHMM", "MultinomialHMM"]



import os
import unittest
from unittest import skipUnless

import pandas as pd
from pandas.testing import assert_frame_equal
from tests.config import test_user, test_dbname, test_hostname, test_password, test_port

from dsbox.dbconnection.dbconnector import DBconnector


@skipUnless(
    os.getenv('WITH_DB_TESTS'),
    'Please set environment variable WITH_DB_TESTS to enable these tests'
)
class TestDBConnector(unittest.TestCase):

    def test_write_df_to_table(self):
        # given
        df = pd.DataFrame({'town': ['Paris', 'New York', 'Roma'],
                           'country': ['France', 'USA', 'Italy'],
                           'trip_cost': [134.6, 234, 85.67]})

        dbconn = DBconnector(username=test_user, password=test_password, hostname=test_hostname, port=test_port,
                             dbname=test_dbname, baseprotocol='postgres://')

        dbconn.engine.execute('DROP TABLE IF EXISTS db.data_town')

        # when
        dbconn.df_to_table(df, 'data_town', schema='db', if_exists='replace', index=False)

        # then
        self.assertTrue(dbconn.check_table('data_town', schema='db'))

    def test_read_table_to_df(self):
        # given
        df_true = pd.DataFrame({'town': ['Paris', 'New York', 'Roma'],
                                'country': ['France', 'USA', 'Italy'],
                                'trip_cost': [134.6, 234, 85.67]})
        dbconn = DBconnector(username=test_user, password=test_password, hostname=test_hostname, port=test_port,
                             dbname=test_dbname, baseprotocol='postgres://')

        # when
        df_returned = dbconn.table_to_df('db.data_town')

        # then
        assert_frame_equal(df_true, df_returned)


if __name__ == '__main__':
    unittest.main()



from sklearn.metrics import f1_score, accuracy_score
from sklearn.ensemble import RandomForestClassifier

from dsbox.utils import write_object_file, load_object_file


def generate_model():
    return RandomForestClassifier(n_estimators=100, n_jobs=-1)


def fit_write_model(dataframe, columns_selection, column_target, write_path, model=generate_model):
    X = dataframe[columns_selection]
    y = dataframe[column_target]

    clf = model()

    clf.fit(X, y)
    write_object_file(write_path, clf)


def read_predict_model(dataframe, columns_selection, read_path, y_pred_column_name='y_prediction'):
    X = dataframe[columns_selection]

    clf = load_object_file(read_path)
    y_pred = clf.predict(X)

    dataframe[y_pred_column_name] = y_pred
    return dataframe


def model_performance(dataframe, y_true_column_name, y_pred_column_name):
    print("Accuracy: " + str(accuracy_score(dataframe[y_true_column_name], dataframe[y_pred_column_name])))
    print("F-score: " + str(f1_score(dataframe[y_true_column_name], dataframe[y_pred_column_name])))



# Copyright 2013 Lars Buitinck

import sys

import numpy as np
from scipy.sparse import csc_matrix

from dsbox.ml.markov.seqlearn.base import BaseSequenceClassifier
from dsbox.ml.markov.seqlearn._utils import (atleast2d_or_csr, check_random_state, count_trans,
                     make_trans_matrix, safe_add, safe_sparse_dot)

__credits__ = "Lars Buitinck"


class StructuredPerceptron(BaseSequenceClassifier):
    """Structured perceptron for sequence classification.

    This implements the averaged structured perceptron algorithm of Collins
    and Daumé, with the addition of an adaptive learning rate.

    Parameters
    ----------
    decode : string, optional
        Decoding algorithm, either "bestfirst" or "viterbi" (default).

    lr_exponent : float, optional
        Exponent for inverse scaling learning rate. The effective learning
        rate is 1. / (t ** lr_exponent), where t is the iteration number.

    max_iter : integer, optional
        Number of iterations (aka. epochs). Each sequence is visited once in
        each iteration.

    random_state : {integer, np.random.RandomState}, optional
        Random state or seed used for shuffling sequences within each
        iteration.

    trans_features : boolean, optional
        Whether to attach features to transitions between labels as well as
        individual labels. This requires more time, more memory and more
        samples to train properly.

    verbose : integer, optional
        Verbosity level. Defaults to zero (quiet mode).

    References
    ----------
    M. Collins (2002). Discriminative training methods for hidden Markov
    models: Theory and experiments with perceptron algorithms. EMNLP.

    Hal Daumé III (2006). Practical Structured Learning Techniques for
    Natural Language Processing. Ph.D. thesis, U. Southern California.
    
    Examples
    --------
    
    >>> import pandas as pd
    >>> from dsbox.ml.markov.seqlearn import StructuredPerceptron
    >>> clf = StructuredPerceptron(max_iter=10)
    >>> X_train = pd.DataFrame({'a': [0, 1, 2], 'b':[0, 1, 2], 'c': [0, 1, 2]})
    >>> y_train = np.array([0, 1, 2])
    >>> X_test = pd.DataFrame({'a': [1, 1, 1], 'b':[1, 1, 1], 'c': [1, 1, 1]})
    >>> _ = clf.fit(X_train, y_train, [len(y_train)])
    >>> clf.predict(X_test)
    array([2, 1, 2])

    """

    def __init__(self, decode="viterbi", lr_exponent=.1, max_iter=10,
                 random_state=None, trans_features=False, verbose=0):
        self.decode = decode
        self.lr_exponent = lr_exponent
        self.max_iter = max_iter
        self.random_state = random_state
        self.trans_features = trans_features
        self.verbose = verbose

    def fit(self, X, y, lengths):
        """Fit to a set of sequences.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Feature matrix of individual samples.

        y : array-like, shape (n_samples,)
            Target labels.

        lengths : array-like of integers, shape (n_sequences,)
            Lengths of the individual sequences in X, y. The sum of these
            should be n_samples.

        Returns
        -------
        self : StructuredPerceptron
        """

        decode = self._get_decoder()

        X = atleast2d_or_csr(X)

        classes, y = np.unique(y, return_inverse=True)
        n_classes = len(classes)

        if self.trans_features:
            trans_true = make_trans_matrix(y, n_classes)
        class_range = np.arange(n_classes)
        Y_true = y.reshape(-1, 1) == class_range

        lengths = np.asarray(lengths)
        n_samples, n_features = X.shape

        end = np.cumsum(lengths)
        start = end - lengths

        w = np.zeros((n_classes, n_features), order='F')
        b_trans = np.zeros((n_classes, n_classes))
        b_init = np.zeros(n_classes)
        b_final = np.zeros(n_classes)

        w_avg = np.zeros_like(w)
        b_trans_avg = np.zeros_like(b_trans)
        b_init_avg = np.zeros_like(b_init)
        b_final_avg = np.zeros_like(b_final)

        if self.trans_features:
            w_trans = np.zeros((n_classes, n_classes, n_features), order='F')
            w_t_flat = w_trans.reshape(-1, n_features)
            w_trans_avg = np.zeros_like(w_trans)
            w_t_avg_flat = w_trans_avg.reshape(-1, n_features)

        sequence_ids = np.arange(lengths.shape[0])
        rng = check_random_state(self.random_state)

        avg_count = 1.
        lr_exponent = self.lr_exponent

        for it in range(1, self.max_iter + 1):
            lr = 1. / (it ** lr_exponent)

            if self.verbose:
                print("Iteration {0:2d}".format(it), end="... ")
                sys.stdout.flush()

            rng.shuffle(sequence_ids)

            sum_loss = 0

            for i in sequence_ids:
                X_i = X[start[i]:end[i]]
                score = safe_sparse_dot(X_i, w.T)
                if self.trans_features:
                    trans_score = safe_sparse_dot(X_i, w_t_flat.T)
                    trans_score = trans_score.reshape(-1, n_classes, n_classes)
                else:
                    trans_score = None
                y_pred = decode(score, trans_score, b_trans, b_init, b_final)
                y_t_i = y[start[i]:end[i]]
                loss = (y_pred != y_t_i).sum()

                if loss:
                    sum_loss += loss

                    Y_t_i = Y_true[start[i]:end[i]]
                    Y_pred = y_pred.reshape(-1, 1) == class_range
                    Y_pred = Y_pred.astype(np.float64)

                    Y_diff = csc_matrix(Y_pred - Y_t_i)
                    Y_diff *= -lr
                    w_update = safe_sparse_dot(Y_diff.T, X_i)

                    if self.trans_features:
                        trans_pred = make_trans_matrix(y_pred, n_classes)

                        trans_diff = trans_true[start[i]:end[i]].toarray()
                        safe_add(trans_diff, -trans_pred)
                        trans_diff = csc_matrix(trans_diff)
                        trans_diff *= -lr

                        w_trans_update = safe_sparse_dot(trans_diff.T, X_i)
                        safe_add(w_t_flat, w_trans_update)

                        w_trans_update *= avg_count
                        safe_add(w_t_avg_flat, w_trans_update)

                    t_trans = count_trans(y_t_i, n_classes)
                    p_trans = count_trans(y_pred, n_classes)
                    b_trans_update = lr * (p_trans - t_trans)
                    b_init_update = lr * (Y_pred[0] - Y_true[start[i]])
                    b_final_update = lr * (Y_pred[-1] - Y_true[end[i] - 1])

                    safe_add(w, w_update)
                    b_trans -= b_trans_update
                    b_init -= b_init_update
                    b_final -= b_final_update

                    w_update *= avg_count
                    b_trans_update *= avg_count
                    b_init_update *= avg_count
                    b_final_update *= avg_count

                    safe_add(w_avg, w_update)
                    b_trans_avg -= b_trans_update
                    b_init_avg -= b_init_update
                    b_final_avg -= b_final_update

            if self.verbose:
                # XXX the loss reported is that for w, but the one for
                # w_avg is what matters for early stopping.
                print("loss = {0:.4f}".format(sum_loss / n_samples))

            avg_count += 1.

        w -= w_avg / avg_count
        if self.trans_features:
            w_trans -= w_trans_avg / avg_count
        b_init -= b_init_avg / avg_count
        b_trans -= b_trans_avg / avg_count
        b_final -= b_final_avg / avg_count

        self.coef_ = w
        if self.trans_features:
            self.coef_trans_ = w_trans
        self.intercept_init_ = b_init
        self.intercept_trans_ = b_trans
        self.intercept_final_ = b_final

        self.classes_ = classes

        return self



import unittest

import numpy as np
import pandas as pd

from dsbox.ml.feature_engineering.timeseries import Shifter, RollingWindower, DistributionTransformer, \
    np_rolling_agg_window, np_rolling_agg_abs_deviation_window
from dsbox.ml.outliers import median_absolute_deviation
from pandas.util.testing import assert_frame_equal, assert_numpy_array_equal, assert_series_equal


class TestComputeShift(unittest.TestCase):
    def test_shift_ts(self):
        # given
        df_ts = pd.DataFrame({'data': [0.0, 1.0, 2.0, 3.0, 4.0]},
                             index=[pd.Timestamp('20130101 09:00:00'),
                                    pd.Timestamp('20130101 09:00:02'),
                                    pd.Timestamp('20130101 09:00:03'),
                                    pd.Timestamp('20130101 09:00:05'),
                                    pd.Timestamp('20130101 09:00:06')])

        # when
        shifter = Shifter(shifts=[1, 2], prefix='t-')
        df_shifted = shifter.transform(df_ts)

        # then
        df_expected = pd.DataFrame(
            {'t-1_data': [np.nan, 0, 1.0, 2.0, 3.0],
             't-2_data': [np.nan, np.nan, 0.0, 1.0, 2.0]},
            index=[pd.Timestamp('20130101 09:00:00'),
                   pd.Timestamp('20130101 09:00:02'),
                   pd.Timestamp('20130101 09:00:03'),
                   pd.Timestamp('20130101 09:00:05'),
                   pd.Timestamp('20130101 09:00:06')]
        )

        self.assertTrue(df_expected.equals(df_shifted))

    def test_shift_several_features(self):
        # given
        df = pd.DataFrame({'x': [0.0, 1.0, 2.0, 3.0, 4.0],
                           'y': [0.0, 0.0, 2.0, 2.0, 1.0]})

        # when
        shifter = Shifter(shifts=[1, 2])
        df_shifted = shifter.transform(df)

        # then
        df_expected = pd.DataFrame(
            {'x_1': [np.nan, 0.0, 1.0, 2.0, 3.0],
             'x_2': [np.nan, np.nan, 0.0, 1.0, 2.0],
             'y_1': [np.nan, 0.0, 0.0, 2.0, 2.0],
             'y_2': [np.nan, np.nan, 0.0, 0.0, 2.0],
             }
        )

        self.assertTrue(df_expected.equals(df_shifted[df_expected.columns]))

    def test_shift_groupby(self):
        # given
        df = pd.DataFrame({'data': [0.0, 1.0, 2.0, 3.0, 4.0],
                           'category': ['A', 'A', 'A', 'B', 'B']})

        # when
        shifter = Shifter(shifts=[1, 2])
        df_shifted = shifter.transform(df.groupby('category'))

        # then
        df_expected = pd.DataFrame(
            {'data_1': [np.nan, 0.0, 1.0, np.nan, 3.0],
             'data_2': [np.nan, np.nan, 0.0, np.nan, np.nan]
             }
        )

        self.assertTrue(df_expected.equals(df_shifted[df_expected.columns]))

    def test_shift_empty_df(self):
        # given
        df = pd.DataFrame()

        # when
        shifter = Shifter()
        df_shifted = shifter.transform(df)

        # then
        df_expected = pd.DataFrame()

        self.assertTrue(df_expected.equals(df_shifted[df_expected.columns]))


class TestRollingWindower(unittest.TestCase):
    def test_compute_moving_average(self):
        df = pd.DataFrame({'data': [0, 1, 2, 3, 4], 'data_bis': [3, 1, 0, 4, 1]})

        roller = RollingWindower(windows=[3], min_periods=1)

        # compute potential differences -> test will fail if a difference is present
        df_mean = pd.DataFrame({'mean_3_data': [0, 0.5, 1, 2, 3], 'mean_3_data_bis': [3, 2, 4 / 3.0, 5 / 3.0, 5 / 3.0]})
        df_test = df_mean.sub(roller.transform(df))
        self.assertEqual(df_test.sum().sum(), 0)

    def test_compute_custom_function(self):
        df = pd.DataFrame({'data': [0, 1, 2, 3, 4], 'data_bis': [3, 1, 0, 4, 1]})

        roller = RollingWindower(operation=median_absolute_deviation, windows=[3], min_periods=1)

        # compute potential differences -> test will fail if a difference is present
        df_mad_true = pd.DataFrame({'mad_3_data': [0, 0.5, 1, 1, 1], 'mad_3_data_bis': [0, 1, 1, 1, 1]})
        df_mad = roller.transform(df)
        df_test = df_mad_true.sub(df_mad)
        self.assertEqual(df_test.sum().sum(), 0)

    def test_compute_moving_average_with_datetime(self):
        df_ts = pd.DataFrame({'data': [0, 1, 2, 3, 4]},
                             index=[pd.Timestamp('20130101 09:00:00'),
                                    pd.Timestamp('20130101 09:00:02'),
                                    pd.Timestamp('20130101 09:00:03'),
                                    pd.Timestamp('20130101 09:00:05'),
                                    pd.Timestamp('20130101 09:00:06')])

        df_mean_ts = pd.DataFrame(
            {'mean_5s_data': [0, 0.5, 1, 2, 2.5],
             'mean_2s_data': [0, 1, 1.5, 3, 3.5]},
            index=[pd.Timestamp('20130101 09:00:00'),
                   pd.Timestamp('20130101 09:00:02'),
                   pd.Timestamp('20130101 09:00:03'),
                   pd.Timestamp('20130101 09:00:05'),
                   pd.Timestamp('20130101 09:00:06')]
        )

        roller = RollingWindower(windows=['5s', '2s'])

        df_test = df_mean_ts.sub(roller.transform(df_ts))
        self.assertEqual(df_test.sum().sum(), 0)


class TestDistribution(unittest.TestCase):
    def test_distribution_transformer_should_produce_df_distribution_per_column(self):
        # given
        df = pd.DataFrame({'sales': [3, 10, 12, 23, 48, 19, 21]})

        # when
        distrib_transformer = DistributionTransformer(3)
        df_distrib = distrib_transformer.fit_transform(df)

        # then
        df_expected = pd.DataFrame({'sales_bin_1': [0, 1, 1, 2, 1, 2, 2],
                                    'sales_bin_2': [0, 0, 0, 0, 1, 0, 0],
                                    'sales_bin_3': [1, 0, 0, 0, 0, 0, 0],
                                    'sales_bin_4': [0, 1, 2, 1, 1, 1, 1]})

        assert_frame_equal(df_expected, df_distrib)

    def test_distribution_transformer_should_produce_df_quantile_per_column(self):
        # given
        df = pd.DataFrame({'sales': [3, 10, 12, 23, 48, 19, 21]})

        # when
        distrib_transformer = DistributionTransformer(3, quantiles=[0., 0.25, 0.5, 0.75, 1.])
        df_distrib = distrib_transformer.fit_transform(df)

        # then
        df_expected = pd.DataFrame({'sales_quantile_1': [3., 3, 3, 10, 12, 19, 19],
                                    'sales_quantile_2': [3, 4.75, 6.5, 11, 17.5, 21, 20],
                                    'sales_quantile_3': [3, 6.5, 10, 12, 23, 23, 21],
                                    'sales_quantile_4': [3, 8.25, 11, 17.5, 35.5, 35.5, 34.5],
                                    'sales_quantile_5': [3., 10, 12, 23, 48, 48, 48],
                                    })

        assert_frame_equal(df_expected, df_distrib)


class TestNumpyRolling(unittest.TestCase):
    def test_np_rolling_mean_window_on_np_array(self):
        # given
        array = np.array([0, 1, 2, 3, 4])

        # when
        array_rol_mean = np_rolling_agg_window(array)

        # then
        array_expected = np.array([np.nan, np.nan, 1, 2, 3])

        assert_numpy_array_equal(array_expected, array_rol_mean)

    def test_np_rolling_median_window_on_np_array(self):
        # given
        array = np.array([0, 1, 2, 3, 0])

        # when
        array_rol_median = np_rolling_agg_window(array, agg_func=np.nanmedian)

        # then
        array_expected = np.array([np.nan, np.nan, 1, 2, 2])

        assert_numpy_array_equal(array_expected, array_rol_median)

    def test_np_rolling_mean_window_on_pd_dataframe(self):
        # given
        df = pd.DataFrame({'values': [0, 1, 2, 3, 4]})

        # when
        df_rol_mean = df.apply(np_rolling_agg_window)

        # then
        df_expected = pd.DataFrame({'values': [np.nan, np.nan, 1, 2, 3]})

        assert_frame_equal(df_expected, df_rol_mean)

    def test_np_rolling_mad_window_on_np_array(self):
        # given
        array = np.array([0, 1, 2, 3, 0])

        # when
        array_rol_mad = np.round(np_rolling_agg_abs_deviation_window(array), 2)

        # then
        array_expected = np.array([np.nan, np.nan, 0.67, 0.67, 1.11])

        assert_numpy_array_equal(array_expected, array_rol_mad)

    def test_np_rolling_mad_window_on_pd_dataframe(self):
        # given
        df = pd.DataFrame({'values': [0, 1, 2, 3, 0]})

        # when
        df_rol_mad = df.apply(np_rolling_agg_abs_deviation_window).round(2)

        # then
        df_expected = pd.DataFrame({'values': [np.nan, np.nan, 0.67, 0.67, 1.11]})

        assert_frame_equal(df_expected, df_rol_mad)

    def test_np_rolling_mad_window_on_pd_dataframe_using_groupby(self):
        # given
        df = pd.DataFrame({
            'group': ['A'] * 5 + ['B'] * 5,
            'values': [0, 1, 2, 3, 4, 0, 1, 2, 3, 0]})

        # when
        serie_rol_group_mad = df.groupby('group')['values'].apply(
            np_rolling_agg_abs_deviation_window).explode().reset_index(drop=True).apply(np.round, decimals=2)

        # then
        serie_expected = pd.Series(name='values',
                                   data=[np.nan, np.nan, 0.67, 0.67, 0.67, np.nan, np.nan, 0.67, 0.67, 1.11])

        assert_series_equal(serie_expected, serie_rol_group_mad)



import unittest

import numpy as np
import pandas as pd
from dsbox.ml.calibration import BinaryCalibrator
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split


class BinaryCalibratorTest(unittest.TestCase):
    def test_params(self):
        X = pd.DataFrame(np.arange(0, 100))
        y = np.zeros(100)
        y[2] = 1
        y[3] = 1

        bin_calib = BinaryCalibrator(RandomForestClassifier(n_estimators=10))
        bin_calib.fit(X, y)

        self.assertEqual(bin_calib.major_class_amount_, 98)
        self.assertEqual(bin_calib.minor_class_amount_, 2)
        self.assertEqual(bin_calib.beta_, 0.020408163265306124)
        self.assertEqual(bin_calib.tau_s_, 0.5)
        self.assertEqual(bin_calib.bias_corrected_threshold(), 0.02)

    def test_undersampling(self):
        X = pd.DataFrame(np.arange(0, 10))
        y = np.zeros(10)
        y[2] = 1
        y[3] = 1

        bin_calib = BinaryCalibrator(RandomForestClassifier(n_estimators=10))
        bin_calib.fit(X, y)

        self.assertEqual(len(bin_calib.X_), 4)

    def test_estimator(self):
        # set random seed for test
        seed = 42
        np.random.seed(seed)

        X, y = make_classification(n_samples=30, n_features=2, n_redundant=0, n_informative=2,
                                   random_state=seed, weights=[0.95, 0.05])

        X_train, X_test, y_train, y_test = \
            train_test_split(X, y, test_size=.2, random_state=seed)

        X_train = pd.DataFrame(X_train)
        X_test = pd.DataFrame(X_test)

        bin_calib = BinaryCalibrator(RandomForestClassifier(random_state=seed, n_estimators=10))
        bin_calib.fit(X_train, y_train)

        y_pred = bin_calib.predict_proba(X_test)

        y_to_pred = np.asmatrix([[0.90789474, 0.09210526],
                                 [0.97183099, 0.02816901],
                                 [0.99519231, 0.00480769],
                                 [0.97183099, 0.02816901],
                                 [0.97183099, 0.02816901],
                                 [0.97183099, 0.02816901]])

        y_pred = np.round(y_pred, 6)
        y_to_pred = np.round(y_to_pred, 6)

        diff_pred = y_pred - y_to_pred

        self.assertEqual(np.sum(diff_pred), 0.0)

        y_pred = bin_calib.predict(X_test)
        self.assertEqual(np.sum(y_pred), 0.0)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ds-box",
    version='1.4.5',
    author='Vincent Levorato',
    url='https://github.com/vlevorato/dsbox',
    description="Package made to accelerate data science workflows in production.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import numpy as np

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


def mean_absolute_deviation(x):
    """
    Mean Absolute Deviation
    """
    return np.nanmean(np.abs(x - np.nanmean(x)))


def median_absolute_deviation(x):
    """
    Median Absolute Deviation
    """
    return np.nanmedian(np.abs(x - np.nanmedian(x)))


def double_median_absolute_deviation(x):
    """
    If the underlying distribution of the data is unsymmetric, one should use the double median 
    absolute deviation, instead of the classic mad.
    
    Parameters
    ----------
    x : array-like

    Returns
    -------
    Couple of values: left mad, right mad

    """
    median = np.nanmedian(x)
    absolute_deviation = np.abs(x - median)

    left_mad = np.nanmedian(absolute_deviation[absolute_deviation <= median])
    right_mad = np.nanmedian(absolute_deviation[absolute_deviation >= median])

    return left_mad, right_mad


def mad_outliers(x, cutoff=2, z_score_coeff=0.6745):
    """
    Median Absolute Deviation outliers method.
    
    It returns values as outliers if their residuals relative to the median exceeds 'n' times the ratio
    to the MAD. The cutoff parameter represents this 'n' times quantity.
    
    Main code: x_mad = np.abs(x - np.median(x)) / mad_value
    if x_mad contains a value above the cutoff parameter, it is considered as an outlier.
    
    Parameters
    ----------
    x : array-like
    
    cutoff : int, optional (default=2)
        amount of times residuals relative to the median exceed the ratio to the MAD
        
    z_score_coeff : float, optional (default=0.6745)
        0.6745 is the 0.75th quartile of the standard normal distribution, to which the MAD converges to.
    
    Returns
    -------
    Boolean array with outlier tag

    """
    mad_value = median_absolute_deviation(x)
    if mad_value == 0:
        return x * np.nan

    x_mad = (z_score_coeff * np.abs(x - np.nanmedian(x))) / mad_value
    return x_mad > cutoff


def fft_outliers(x, freq_cut_index=None, outlier_proportion=0.1):
    """
    Fast Fourier Transformation outliers method.
    
    The idea is to isolate high frequencies in frequency domain, i.e. fast variations. Isolating such frequencies
    allows to identify abrupt change in numerical series, which are considered here as outliers.
    
    Parameters
    ----------
    x : array-like
    
    freq_cut_index : int, optional (default=None)
        set the index in the FFT numpy array beyond which the frequency is considered to be filtered. The highest,
        the less filtering. By default, it sets this index to 0.9 of x length
        
    outlier_proportion : float, optional (default=0.1)
        set the proportion of outliers to return

    Returns
    -------
    Boolean array with outlier tag
    
    """
    if freq_cut_index is None:
        freq_cut_index = int(0.9 * float(len(x)))

    fft_x = np.fft.fft(x)
    for freq_index in range(0, len(fft_x)):
        if freq_index >= freq_cut_index:
            fft_x[freq_index] = 0

    ifft_x = np.real(np.fft.ifft(fft_x).round())
    diff_fft_x = np.abs(x - ifft_x)
    outlier_amount = int(len(x) * outlier_proportion)
    outliers = np.zeros(len(x), dtype='bool')
    for i in range(0, outlier_amount):
        position_outlier = np.argmax(diff_fft_x)
        outliers[position_outlier] = True
        diff_fft_x[position_outlier] = -1

    return outliers



import google.auth
from google.cloud import bigquery
from google.cloud import bigquery_storage

SCOPES = ["https://www.googleapis.com/auth/cloud-platform", "https://www.googleapis.com/auth/drive"]


def get_bq_client():
    """
    Build a BigQuery Python client.

    Returns
    -------
    bigquery.Client instance
    """
    credentials, project_id = google.auth.default(
        scopes=SCOPES
    )

    bqclient = bigquery.Client(
        credentials=credentials,
        project=project_id,
    )

    return bqclient


def get_bq_storage_client():
    """
    Build a BigQueryStorage Python client.

    Returns
    -------
    bigquery_storage.BigQueryReadClient instance
    """
    credentials, project_id = google.auth.default(
        scopes=SCOPES
    )

    bqstorageclient = bigquery_storage.BigQueryReadClient(
        credentials=credentials
    )

    return bqstorageclient



import numpy as np
from scipy import linalg


def log_multivariate_normal_density(X, means, covars, covariance_type='diag'):
    """Compute the log probability under a multivariate Gaussian distribution.
    Parameters
    ----------
    X : array_like, shape (n_samples, n_features)
        List of n_features-dimensional data points. Each row corresponds to a
        single data point.
    means : array_like, shape (n_components, n_features)
        List of n_features-dimensional mean vectors for n_components Gaussians.
        Each row corresponds to a single mean vector.
    covars : array_like
        List of n_components covariance parameters for each Gaussian. The shape
        depends on `covariance_type`:
            (n_components, n_features)      if 'spherical',
            (n_features, n_features)    if 'tied',
            (n_components, n_features)    if 'diag',
            (n_components, n_features, n_features) if 'full'
    covariance_type : string
        Type of the covariance parameters.  Must be one of
        'spherical', 'tied', 'diag', 'full'.  Defaults to 'diag'.
    Returns
    -------
    lpr : array_like, shape (n_samples, n_components)
        Array containing the log probabilities of each data point in
        X under each of the n_components multivariate Gaussian distributions.
    """
    log_multivariate_normal_density_dict = {
        'spherical': _log_multivariate_normal_density_spherical,
        'tied': _log_multivariate_normal_density_tied,
        'diag': _log_multivariate_normal_density_diag,
        'full': _log_multivariate_normal_density_full}
    return log_multivariate_normal_density_dict[covariance_type](
        X, means, covars
    )


def _log_multivariate_normal_density_diag(X, means, covars):
    """Compute Gaussian log-density at X for a diagonal model."""
    # X: (ns, nf); means: (nc, nf); covars: (nc, nf) -> (ns, nc)
    n_samples, n_dim = X.shape
    # Avoid 0 log 0 = nan in degenerate covariance case.
    covars = np.maximum(covars, np.finfo(float).tiny)
    with np.errstate(over="ignore"):
        return -0.5 * (n_dim * np.log(2 * np.pi)
                       + np.log(covars).sum(axis=-1)
                       + ((X[:, None, :] - means) ** 2 / covars).sum(axis=-1))


def _log_multivariate_normal_density_spherical(X, means, covars):
    """Compute Gaussian log-density at X for a spherical model."""
    cv = covars.copy()
    if covars.ndim == 1:
        cv = cv[:, np.newaxis]
    if cv.shape[1] == 1:
        cv = np.tile(cv, (1, X.shape[-1]))
    return _log_multivariate_normal_density_diag(X, means, cv)


def _log_multivariate_normal_density_tied(X, means, covars):
    """Compute Gaussian log-density at X for a tied model."""
    cv = np.tile(covars, (means.shape[0], 1, 1))
    return _log_multivariate_normal_density_full(X, means, cv)


def _log_multivariate_normal_density_full(X, means, covars, min_covar=1.e-7):
    """Log probability for full covariance matrices."""
    n_samples, n_dim = X.shape
    nmix = len(means)
    log_prob = np.empty((n_samples, nmix))
    for c, (mu, cv) in enumerate(zip(means, covars)):
        try:
            cv_chol = linalg.cholesky(cv, lower=True)
        except linalg.LinAlgError:
            # The model is most probably stuck in a component with too
            # few observations, we need to reinitialize this components
            try:
                cv_chol = linalg.cholesky(cv + min_covar * np.eye(n_dim),
                                          lower=True)
            except linalg.LinAlgError:
                raise ValueError("'covars' must be symmetric, "
                                 "positive-definite")

        cv_log_det = 2 * np.sum(np.log(np.diagonal(cv_chol)))
        cv_sol = linalg.solve_triangular(cv_chol, (X - mu).T, lower=True).T
        log_prob[:, c] = - .5 * (np.sum(cv_sol ** 2, axis=1) +
                                 n_dim * np.log(2 * np.pi) + cv_log_det)

    return log_prob



import unittest

from dsbox.ml.text.utils import shinglelize


class TestTextFeatureExtraction(unittest.TestCase):
    def test_shinglelize_function(self):
        # given
        text = "There is a better way."

        # when
        shingles_obtained = shinglelize(text)

        # then
        expected_shingle_list = ['The', 'her', 'ere', 're ', 'e i', ' is', 'is ', 's a', ' a ', 'a b', ' be', 'bet',
                                 'ett', 'tte', 'ter', 'er ', 'r w', ' wa', 'way', 'ay.']
        self.assertEqual(expected_shingle_list, shingles_obtained)

    def test_shinglelize_function_on_empty_string(self):
        # given
        text = ""

        # when
        shingles_obtained = shinglelize(text)

        # then
        self.assertEqual(shingles_obtained, [])


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime

from airflow import DAG
from airflow.models import TaskInstance

from dsbox.dbconnection.plasma_connector import PlasmaConnector
from dsbox.operators.data_operator import DataOperator
from dsbox.operators.data_unit import DataInputFileUnit, DataOutputPlasmaUnit

from tests.config import socket_name, object_id


def drop_na_dataframe(dataframe, columns):
    dataframe = dataframe.dropna(subset=columns)
    return dataframe


class TestPlasmaConnector(unittest.TestCase):
    def test_execute_data_operator_csv_read_and_plasma_write(self):
        # given
        plasma_connector = PlasmaConnector(socket_name)

        dag = DAG(dag_id='test', start_date=datetime.now())
        input_csv_unit = DataInputFileUnit('data/X.csv', sep=';')
        output_plasma_unit = DataOutputPlasmaUnit(plasma_connector, object_id)

        task = DataOperator(operation_function=drop_na_dataframe,
                            params={'columns': ['ANNEEREALISATIONDIAGNOSTIC']},
                            input_unit=input_csv_unit,
                            output_unit=output_plasma_unit,
                            dag=dag, task_id='data_operator_csv_to_parquet')

        task_instance = TaskInstance(task=task, execution_date=datetime.now())

        # when
        task.execute(task_instance.get_template_context())

        # then
        other_plasma_connector = PlasmaConnector(socket_name)
        df_transformed = other_plasma_connector.get_dataframe(object_id)
        self.assertEqual((10245, 27), df_transformed.shape)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as st
from mpl_toolkits.mplot3d import Axes3D

__author__ = "Vincent Levorato"
__credits__ = "Madalina Ciortan"
__source__ = "https://towardsdatascience.com/simple-example-of-2d-density-plots-in-python-83b83b934f67"
__license__ = "Apache 2.0"


def plot_gaussian_density(X, dimensions=2):
    """
    Plot a 2d gaussian 
    
    Parameters
    ----------
    X: array-like (2D)
        data used to estimate density
    
    dimensions: int, optional, (default=2)
        plot dimension of gaussian densities
    
    Examples
    --------
    
    >>> from dsbox.ml.visualization.distribution import plot_gaussian_density
    >>> from sklearn.datasets import make_blobs
    
    >>> n_components = 3
    >>> X , _ = make_blobs(n_samples=300, centers=n_components, cluster_std = [2, 1.5, 1], random_state=42)
    
    >>> plot_gaussian_density(X, dimensions=2)
    >>> plot_gaussian_density(X, dimensions=3)

    """

    # Extract x and y
    x = X[:, 0]
    y = X[:, 1]
    # Define the borders
    deltaX = (np.max(x) - np.min(x)) / 10
    deltaY = (np.max(y) - np.min(y)) / 10
    xmin = np.min(x) - deltaX
    xmax = np.max(x) + deltaX
    ymin = np.min(y) - deltaY
    ymax = np.max(y) + deltaY

    # Create meshgrid
    xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]

    positions = np.vstack([xx.ravel(), yy.ravel()])
    values = np.vstack([x, y])
    kernel = st.gaussian_kde(values)
    f = np.reshape(kernel(positions).T, xx.shape)

    if dimensions == 2:
        fig = plt.figure(figsize=(8, 8))
        ax = fig.gca()
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        cfset = ax.contourf(xx, yy, f, cmap='coolwarm')
        ax.imshow(np.rot90(f), cmap='coolwarm', extent=[xmin, xmax, ymin, ymax])
        cset = ax.contour(xx, yy, f, colors='k')
        ax.clabel(cset, inline=1, fontsize=10)
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        plt.title('2D Gaussian Kernel density estimation')
        plt.show()

    if dimensions == 3:
        fig = plt.figure(figsize=(13, 7))
        ax = Axes3D(fig)
        surf = ax.plot_surface(xx, yy, f, rstride=1, cstride=1, cmap='coolwarm', edgecolor='none')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('PDF')
        ax.set_title('Surface plot of Gaussian 2D KDE')
        fig.colorbar(surf, shrink=0.5, aspect=5)  # add color bar indicating the PDF
        ax.view_init(60, 35)
        plt.show()



import numpy as np
import pandas as pd
from numpy.lib.stride_tricks import as_strided
from sklearn.base import BaseEstimator, TransformerMixin
import types

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class Shifter(BaseEstimator, TransformerMixin):
    """
        Compute shifted values for a given dataframe, and creates columns associated.

        Parameters
        ----------   

            shifts: list, optional, default [1]
                Allows to shift data by periods (meaning by row) using the pandas.shift() method.

            prefix: string, optional
                Put a prefix in front of the column names.

            suffix: string, optional
                Put a suffix at the end of the column names.

            group_by_cols: list, optional, default None
                Allows to shift values grouped by columns.

            ignore_cols: list, optional, default None
                Allows to ignore some columns to be not shifted.

        Examples
        --------

        >>> from dsbox.ml.feature_engineering.timeseries import Shifter
        >>> df_ts = pd.DataFrame({'data': [0.0, 1.0, 2.0, 3.0, 4.0]})
        >>> shifter = Shifter(shifts=[1],prefix='t-')
        >>> df_shift_ts = shifter.transform(df_ts)
        >>> df_shift_ts    
           t-1_data
        0       NaN
        1       0.0
        2       1.0
        3       2.0
        4       3.0


    """

    def __init__(self, shifts=[1], prefix='', suffix='', **kwargs):
        self.shifts = shifts
        self.prefix = prefix
        self.suffix = suffix
        self.kwargs = kwargs

    def fit(self, X=None, y=None):
        """
        No-op.
        This method doesn't do anything. It exists purely for compatibility
        with the scikit-learn transformer API.

        Parameters
        ----------
            X: array-like
            y: array-like

        Returns
        -------
            self: Shifter

        """

        return self

    def transform(self, X):
        """
        Transform a dataframe into shift values corresponding to shift periods

        Parameters
        ----------
            X: dataframe
                Input pandas dataframe.

        Returns
        -------
            X: dataframe 
                Dataframe with columns having shifted values (one by shift)

        """

        X_shifted = pd.DataFrame()

        for shift_value in self.shifts:
            X_shifted_tmp = X
            X_shifted_tmp = X_shifted_tmp.shift(periods=shift_value, **self.kwargs)

            prefix = ''
            suffix = ''
            if self.prefix != '':
                prefix = self.prefix + str(shift_value) + '_'
            if self.suffix != '':
                suffix = self.suffix + str(shift_value)
            if self.suffix == '' and self.prefix == '':
                suffix = '_' + str(shift_value)

            X_shifted_tmp.columns = X_shifted_tmp.columns.map(lambda x: prefix + x + suffix)
            if len(X_shifted) == 0:
                X_shifted = X_shifted_tmp
            else:
                X_shifted = X_shifted.join(X_shifted_tmp)

        return X_shifted


class RollingWindower(BaseEstimator, TransformerMixin):
    """
    Compute rolling aggregated values for a given dataframe.

    Classical operators (mean, std, median, etc.) or custom operators can be used to compute the windows. 
    Windows are based on index, which could be a simple integer or a pandas timestamp. Use **kwargs to pass extra
    arguments to pandas rolling function (like min_periods for instance).


    Parameters
    ----------   
        operation: str or function, optional, default 'mean'
            Set the aggregation function used to aggregate a window.

        windows: list, optional, default [3]
            Set the windows used to aggregate data. Time windows can be set also (see pandas time unit
            syntax) if the dataframe index is a timestamp
            Examples:
                [3]: one window of size 3
                [2,5]: one window of size 2 and one window of size 5
                [2s, 10s]: one window of 2 secs and one window of 10 secs

        diff_mode: boolean, optional, default False
            Process the difference between values and its window aggregate value.


    Examples
    --------
        >>> import pandas as pd
        >>> from dsbox.ml.feature_engineering.timeseries import RollingWindower
        >>> df = pd.DataFrame({'data': [0, 1, 2, 3, 4]})
        >>> roller = RollingWindower(windows=[2,3])
        >>> df_roll = roller.transform(df)
        >>> df_roll
           mean_2_data  mean_3_data
        0          NaN          NaN
        1          0.5          NaN
        2          1.5          1.0
        3          2.5          2.0
        4          3.5          3.0

        >>> df_ts = pd.DataFrame({'data': [0, 1, 2, 3, 4]}, \
                            index=[pd.Timestamp('20130101 09:00:00'), \
                                pd.Timestamp('20130101 09:00:02'), \
                                pd.Timestamp('20130101 09:00:03'), \
                                pd.Timestamp('20130101 09:00:05'), \
                                pd.Timestamp('20130101 09:00:06')])
        >>> roller = RollingWindower(windows=['5s', '2s'])
        >>> df_roll = roller.transform(df_ts)
        >>> df_roll
                             mean_5s_data  mean_2s_data
        2013-01-01 09:00:00           0.0           0.0
        2013-01-01 09:00:02           0.5           1.0
        2013-01-01 09:00:03           1.0           1.5
        2013-01-01 09:00:05           2.0           3.0
        2013-01-01 09:00:06           2.5           3.5



    """

    def __init__(self, operation='mean', windows=[3], **kwargs):
        self.operation = operation
        self.windows = windows
        self.kwargs = kwargs

    def fit(self, X=None, y=None):
        """
        No-op.
        This method doesn't do anything. It exists purely for compatibility
        with the scikit-learn transformer API.

        Parameters
        ----------
            X: array-like
            y: array-like

        Returns
        -------
            self: RollingWindower

        """

        return self

    def transform(self, raw_X):
        """
        Transform a dataframe into aggregated values corresponding to window sizes

        Parameters
        ----------
            raw_X: dataframe
                Input pandas dataframe.

        Returns
        -------
            X: dataframe 
                Dataframe with columns having rolling measures (one per window)

        """

        X = pd.DataFrame()

        for window in self.windows:

            X_m = raw_X.rolling(window, **self.kwargs).agg(self.operation)

            columns_name = []

            if isinstance(self.operation, types.FunctionType):
                col_name = self.operation.__name__
            else:
                col_name = self.operation

            for col in X_m.columns:
                columns_name.append(col_name + '_' + str(window) + '_' + col)

            X_m.columns = columns_name

            if len(X) == 0:
                X = X_m
            else:
                X = X.join(X_m)

        return X


class DistributionTransformer(BaseEstimator, TransformerMixin):
    """
    Build a discrete distribution (histogram) for feature engineering for each column, per line, 
    following a rolling window. It captures the evolving distribution of a feature.

    For instance, if a serie is composed of the following values: [3, 10, 12, 23], and with a window parameter of 3,
    it takes these values, and apply histogram (bins=4) function on it:
    [3] => [0, 0, 1, 0]
    [3, 10] => [1, 0, 0, 1]
    [3, 10, 12] => [1, 0, 0, 2]
    [10, 12, 23] => [2, 0, 0, 1]

    Parameters
    ----------   
        window: int
            Size of the rolling window.

        bins: int, optional, (default=4)
            Amount of bins used to estimate distribution
        
        quantiles: int, optional, (default=None)
            If set, the transformer will return quantiles information.

    Examples
    --------
    >>> import pandas as pd
    >>> from dsbox.ml.feature_engineering.timeseries import DistributionTransformer

    >>> df = pd.DataFrame({'sales': [3, 10, 12, 23, 48, 19, 21]})
    >>> distrib_transformer = DistributionTransformer(3)
    >>> distrib_transformer.fit_transform(df)
       sales_bin_1  sales_bin_2  sales_bin_3  sales_bin_4
    0            0            0            1            0
    1            1            0            0            1
    2            1            0            0            2
    3            2            0            0            1
    4            1            1            0            1
    5            2            0            0            1
    6            2            0            0            1
    """

    def __init__(self, window, bins=4, quantiles=None):
        self.window = window
        self.bins = bins
        self.quantiles = quantiles

    def fit(self, X=None, y=None):
        """
        No-op.
        This method doesn't do anything. It exists purely for compatibility
        with the scikit-learn transformer API.

        Parameters
        ----------
            X: array-like
            y: array-like

        Returns
        -------
            self

        """

        return self

    def transform(self, X):
        """
        Transform a dataframe to build discrete distribution per column.

        Parameters
        ----------
           X: dataframe
               Input pandas dataframe.

        Returns
        -------
           X: dataframe 
               Dataframe with bin values per column.

        """

        X_distrib = pd.DataFrame()

        for col in X.columns:
            col_serie = X[col]
            bins_list = []

            for i in range(0, len(col_serie)):
                min_bound = i - self.window + 1
                if min_bound < 0:
                    min_bound = 0
                max_bound = i + 1
                if max_bound >= len(col_serie):
                    max_bound = len(col_serie)
                if self.quantiles is None:
                    bins_list.append(np.histogram(col_serie[min_bound:max_bound], bins=self.bins)[0])
                else:
                    bins_list.append(np.quantile(col_serie[min_bound:max_bound], self.quantiles))

            X_col_distrib = pd.DataFrame(bins_list)
            X_col_distrib = X_col_distrib.set_index(X.index)
            if self.quantiles is None:
                X_col_distrib.columns = [col + '_bin_' + str(i) for i in range(1, self.bins + 1)]
            else:
                X_col_distrib.columns = [col + '_quantile_' + str(i) for i in range(1, len(self.quantiles) + 1)]
                X_col_distrib = X_col_distrib.fillna(0)

            if len(X_distrib) == 0:
                X_distrib = X_col_distrib
            else:
                X_distrib = X_distrib.join(X_col_distrib)

        return X_distrib


def transform_datetxt2int(X, col_date, format='%Y-%m-%d'):
    """
    Inplace transformation of a string date column into an integer format date column.

    Parameters
    ----------
    X: dataframe

    col_date: str
        Column name to transform

    format: str
        Pandas date str format

    """

    X[col_date] = pd.to_datetime(X[col_date], format=format)
    X[col_date] = X[col_date].map(lambda x: (x.year * 10 ** 4) + (x.month * 10 ** 2) + x.day)
    X[col_date] = X[col_date].astype('int')


def create_diff_shift_features(df_shift, cols=[], prefix='diff_'):
    """
    Create diff values between time series columns with have been shifted.

    Parameters
    ----------
    df_shift: dataframe
        Dataset with columns to make apply diff.
    cols: list
        Columns names, in order, to apply diff.
    prefix: str
        Prefix used to name diff columns.

    """
    for i in range(0, len(cols) - 1):
        df_shift[prefix + cols[i + 1] + '_' + cols[i]] = df_shift[cols[i + 1]] - df_shift[cols[i]]


def _get_numpy_array_strides(array, window):
    shape = array.shape[:-1] + (array.shape[-1], window)
    strides = array.strides + (array.strides[-1],)
    sequence_strides = as_strided(array, shape=shape, strides=strides)
    return sequence_strides


def np_rolling_agg_window(array, window=3, agg_func=np.nanmean):
    """
    Calculating rolling aggregation using function, based on zero memory copying with numpy.
    By default, the aggregation function is mean, so it returns rolling mean.
    Efficient version of Pandas rolling function (with less parameters).

    Parameters
    ----------
    array: numpy.array or pandas.Series
        array values
    window: int, default=3
        window size for the aggregate function rolling calculation
    agg_func: Callable, default=numpy.nanmean
        function used for aggregation

    Returns
    -------
    agg: numpy.array
        Rolling result.


    """
    if type(array) == pd.Series:
        array = array.values

    sequence_strides = _get_numpy_array_strides(array, window)

    agg = np.roll(agg_func(sequence_strides, axis=1), window - 1)
    agg[0:window - 1] = np.nan
    return agg


def np_rolling_agg_abs_deviation_window(array, window=3, agg_func=np.nanmean):
    """
    Calculating rolling absolute deviation related to an aggregation function, using zero memory copying based on numpy.
    By default, the aggregation function is mean, so it returns rollling Mean Absolute Deviation. Other calculation can be done
    like the rolling Median Absolute Deviation.

    Parameters
    ----------
    array: numpy.array or pandas.Series
        array values
    window: int, default=3
        window size for the aggregate function rolling calculation
    agg_func: Callable, default=numpy.nanmean
        function used for aggregation

    Returns
    -------
    agg_abs_deviation: numpy.array
        Rolling "Agg" Absolute Deviation result.


    """
    if type(array) == pd.Series:
        array = array.values

    sequence_strides = _get_numpy_array_strides(array, window)

    m = agg_func(sequence_strides, axis=1).reshape(-1, 1)
    agg_abs_deviation = np.roll(agg_func(np.abs(sequence_strides - m), axis=1), window - 1)
    agg_abs_deviation[0:window - 1] = np.nan
    return agg_abs_deviation



from collections import OrderedDict

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class HistogramData:
    """
    Class which represents the frequency distribution of given data.
    
    """
    def __init__(self, intervals, data):
        self.interval_dict = OrderedDict()  # structure: key (if value > key and value < key+1) => amount
        self.amount_data = 0

        for index, element in enumerate(data):
            self.interval_dict[intervals[index]] = element
            self.amount_data += element

        self.interval_dict[intervals[len(intervals) - 1]] = 0

    def get_amount(self, value, normalized=False, cumulative=None):
        keys = list(self.interval_dict.keys())

        sum_values = 0

        if len(keys) == 0:
            return 0
        if len(keys) == 1:
            amount = 0
            if self.interval_dict.get(keys[0]) is not None:
                amount = self.interval_dict[keys[0]]
                if normalized:
                    if self.amount_data != 0:
                        amount = float(self.interval_dict[keys[0]]) / self.amount_data

            return amount

        found = False
        i = len(keys) - 1

        while not found:
            if value >= keys[i]:
                found = True
            else:
                if self.interval_dict.get(keys[i]) is not None:
                    sum_values += self.interval_dict[keys[i]]
                i -= 1
                if i < 0:
                    return 0

        divisor = 1.0
        if normalized:
            divisor = self.amount_data

        if self.interval_dict.get(keys[i]) is not None:
            if cumulative is None:
                return float(self.interval_dict[keys[i]]) / divisor
            if cumulative == 'desc':
                return float(sum_values) / divisor
            if cumulative == 'asc':
                return float(self.amount_data - sum_values) / divisor
        else:
            return 0



from datetime import datetime
import os

from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from airflow.operators.subdag_operator import SubDagOperator

from dsbox.examples.tree_disease_usecase.ml.feature_engineering import join_dataframes, transform_df_coordinates
from dsbox.examples.tree_disease_usecase.ml.modeling import fit_write_model, read_predict_model, model_performance
from dsbox.examples.tree_disease_usecase.ml.sub_dags import feature_engineering_sub_dag
from dsbox.operators.data_operator import DataOperator
from dsbox.operators.data_unit import DataInputFileUnit, DataOutputFileUnit, DataInputMultiFileUnit, DataOutputDBUnit, \
    DataOutputPGUnit
from dsbox.utils import execute_dag, plot_dag
from dsbox.utils import FilenameGenerator

project_path = os.getenv('PROJECT_PATH')

pg_connection_dict = {}
pg_connection_dict['username'] = os.getenv('DSBOX_PG_USERNAME')
pg_connection_dict['password'] = os.getenv('DSBOX_PG_PASSWORD')
pg_connection_dict['hostname'] = os.getenv('DSBOX_PG_HOSTNAME')
pg_connection_dict['port'] = os.getenv('DSBOX_PG_PORT')
pg_connection_dict['dbname'] = os.getenv('DSBOX_PG_DBNAME')


def dummy_function(dataframe):
    return dataframe.fillna('NA')


features_selection = ['ADR_SECTEUR', 'ANNEEDEPLANTATION', 'coord_x', 'coord_y',
                      'ANNEEREALISATIONDIAGNOSTIC', 'ANNEETRAVAUXPRECONISESDIAG',
                      'GENRE_BOTA', 'ESPECE', 'DIAMETREARBREAUNMETRE', 'FREQUENTATIONCIBLE',
                      'NOTEDIAGNOSTIC', 'PRIORITEDERENOUVELLEMENT', 'RAISONDEPLANTATION', 'REMARQUES',
                      'SOUS_CATEGORIE', 'STADEDEDEVELOPPEMENT', 'STADEDEVELOPPEMENTDIAG',
                      'TRAITEMENTCHENILLES', 'TRAVAUXPRECONISESDIAG', 'TROTTOIR',
                      'VARIETE', 'VIGUEUR', 'CODE_PARENT']

feature_target = 'Defaut'
prediction_column_name = 'y_prediction'

filename_generator = FilenameGenerator(path=project_path + 'datasets/temp/')
temp_files = []
for i in range(0, 100):
    temp_files.append(filename_generator.generate_filename() + '.parquet')

dag = DAG(dag_id='Tree_Disease_Prediction_with_pg', description='Tree Disease Prediction Example (with PG)',
          schedule_interval='0 12 * * *', start_date=datetime(2017, 3, 20), catchup=False)

input_csv_files_unit = DataInputMultiFileUnit([project_path + 'datasets/input/X_tree_egc_t1.csv',
                                               project_path + 'datasets/input/X_geoloc_egc_t1.csv',
                                               project_path + 'datasets/input/Y_tree_egc_t1.csv'], sep=';')
output_parquet_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_train_raw.parquet',
                                         pandas_write_function_name='to_parquet')
task_concate_train_files = DataOperator(operation_function=join_dataframes,
                                        input_unit=input_csv_files_unit,
                                        output_unit=output_parquet_unit,
                                        dag=dag, task_id='Join_train_data_source_files')

task_feature_engineering_for_train = SubDagOperator(
    subdag=feature_engineering_sub_dag(dag.dag_id, 'Feature_engineering_for_train',
                                       model_path=project_path + 'models/',
                                       input_file=project_path + 'datasets/temp/X_train_raw.parquet',
                                       output_file=project_path + 'datasets/temp/X_train_final.parquet',
                                       temp_files=temp_files[0:10],
                                       start_date=dag.start_date,
                                       schedule_interval=dag.schedule_interval),
    task_id='Feature_engineering_for_train',
    dag=dag,
)

task_concate_train_files.set_downstream(task_feature_engineering_for_train)

input_parquet_raw_file_unit = DataInputFileUnit(project_path + 'datasets/temp/X_train_final.parquet',
                                                pandas_read_function_name='read_parquet')
task_model_learning = DataOperator(operation_function=fit_write_model,
                                   params={'columns_selection': features_selection,
                                           'column_target': feature_target,
                                           'write_path': project_path + 'models/ensemble.model'
                                           },
                                   input_unit=input_parquet_raw_file_unit,
                                   dag=dag, task_id='Model_learning')

task_feature_engineering_for_train.set_downstream(task_model_learning)

input_csv_files_unit = DataInputMultiFileUnit([project_path + 'datasets/input/X_tree_egc_t2.csv',
                                               project_path + 'datasets/input/X_geoloc_egc_t2.csv',
                                               project_path + 'datasets/input/Y_tree_egc_t2.csv'], sep=';')
output_parquet_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_test_raw.parquet',
                                         pandas_write_function_name='to_parquet')
task_concate_test_files = DataOperator(operation_function=join_dataframes,
                                       input_unit=input_csv_files_unit,
                                       output_unit=output_parquet_unit,
                                       dag=dag, task_id='Join_test_data_source_files')

task_feature_engineering_for_test = SubDagOperator(
    subdag=feature_engineering_sub_dag(dag.dag_id, 'Feature_engineering_for_test',
                                       model_path=project_path + 'models/',
                                       input_file=project_path + 'datasets/temp/X_test_raw.parquet',
                                       output_file=project_path + 'datasets/temp/X_test_final.parquet',
                                       temp_files=temp_files[10:],
                                       start_date=dag.start_date,
                                       schedule_interval=dag.schedule_interval,
                                       mode='predict'),
    task_id='Feature_engineering_for_test',
    dag=dag,
)

task_concate_test_files.set_downstream(task_feature_engineering_for_test)
task_feature_engineering_for_train.set_downstream(task_feature_engineering_for_test)

task_model_predict = DataOperator(operation_function=read_predict_model,
                                  params={'columns_selection': features_selection,
                                          'read_path': project_path + 'models/ensemble.model',
                                          'y_pred_column_name': prediction_column_name
                                          },
                                  input_unit=DataInputFileUnit(project_path + 'datasets/temp/X_test_final.parquet',
                                                               pandas_read_function_name='read_parquet'),
                                  output_unit=DataOutputFileUnit(project_path + 'datasets/output/X_predict.csv',
                                                                 pandas_write_function_name='to_csv',
                                                                 index=False),
                                  dag=dag, task_id='Model_prediction')

task_feature_engineering_for_test.set_downstream(task_model_predict)
task_model_learning.set_downstream(task_model_predict)

input_result_file_unit = DataInputFileUnit(project_path + 'datasets/output/X_predict.csv',
                                           pandas_read_function_name='read_csv')
task_model_metric = DataOperator(operation_function=model_performance,
                                 params={'y_true_column_name': feature_target,
                                         'y_pred_column_name': prediction_column_name
                                         },
                                 input_unit=input_result_file_unit,
                                 dag=dag, task_id='Model_metrics')

task_model_predict.set_downstream(task_model_metric)

output_result_unit = DataOutputPGUnit(pg_connection_dict, table_name='tree_disease.predictions', to_pg_drop=True)
task_export_to_psql = DataOperator(operation_function=transform_df_coordinates,
                                   params={'x_column': 'coord_x',
                                             'y_column': 'coord_y'
                                             },
                                   input_unit=input_result_file_unit,
                                   output_unit=output_result_unit,
                                   dag=dag, task_id='Export_result_to_Postgres')

task_model_predict.set_downstream(task_export_to_psql)

task_purge_temp_files = BashOperator(task_id='Purge_temp_files',
                                     bash_command='rm ' + project_path + 'datasets/temp/*',
                                     dag=dag)

task_export_to_psql.set_downstream(task_purge_temp_files)



import random
import numpy as np

__author__ = "Vincent Levorato"
__license__ = "Apache 2.0"


def greedy_feature_selection(X_train, X_test, y_train, y_test, model, columns, score_func, shuffle=False):
    """
    Greedy technique to select features in a model. It takes all the columns passed into the function at first, and remove
    feature one by one, recalculating performance each time (train/test/score). If removing a feature increases the score
    it is evinced, else, it is kept.

    Parameters
    ----------
        X_train: pandas.Dataframe
        X_test: pandas.Dataframe
        y_train : array-like
        y_test : array-like
        model: sklearn estimator
        columns : array-like
        score_func : function
        shuffle : bool (optional)

    Returns
    -------
    Return a sublist of features

    """
    cols_selected = list(columns)

    if shuffle:
        random.shuffle(cols_selected)

    i = 0
    stop = False

    model.fit(X_train[cols_selected], y_train)
    y_pred = np.round(model.predict(X_test[cols_selected]))
    score = score_func(y_test, y_pred)
    print(score)
    print('')

    while not stop:
        col_to_remove = cols_selected.pop(0)
        print(col_to_remove)
        model.fit(X_train[cols_selected], y_train)
        y_pred = np.round(model.predict(X_test[cols_selected]))

        score_temp = score_func(y_test, y_pred)
        if score_temp <= score:
            score = score_temp
            print(score)
            i = 0
        else:
            cols_selected.append(col_to_remove)

        i += 1

        if i == len(cols_selected):
            stop = True
            print("No more optim")

    return cols_selected
